#
# Copyright 2015 Benjamin Kiessling
#           2014 Thomas M. Breuel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.pageseg
~~~~~~~~~~~~~~

Layout analysis methods.
"""
import logging
import uuid
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import PIL
from scipy.ndimage import gaussian_filter, maximum_filter, uniform_filter

from kraken.containers import BBoxLine, Segmentation
from kraken.lib import morph, sl
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.segmentation import reading_order, topsort
from kraken.lib.util import get_im_str, is_bitonal, pil2array

__all__ = ['segment']

logger = logging.getLogger(__name__)


class record(object):
    """
    Simple dict-like object.
    """
    def __init__(self, **kw):
        self.__dict__.update(kw)
        self.label: int = 0
        self.bounds: List = []
        self.mask: np.ndarray = None


def find(condition):
    "Return the indices where ravel(condition) is true"
    res, = np.nonzero(np.ravel(condition))
    return res


def binary_objects(binary: np.ndarray) -> np.ndarray:
    """
    Labels features in an array and segments them into objects.
    """
    labels, _ = morph.label(binary)
    objects = morph.find_objects(labels)
    return objects


def estimate_scale(binary: np.ndarray) -> float:
    """
    Estimates image scale based on number of connected components.
    """
    objects = binary_objects(binary)
    bysize = sorted(objects, key=sl.area)
    scalemap = np.zeros(binary.shape)
    for o in bysize:
        if np.amax(scalemap[o]) > 0:
            continue
        scalemap[o] = sl.area(o)**0.5
    scale = np.median(scalemap[(scalemap > 3) & (scalemap < 100)])
    return scale


def compute_boxmap(binary: np.ndarray, scale: float,
                   threshold: Tuple[float, int] = (.5, 4),
                   dtype: str = 'i') -> np.ndarray:
    """
    Returns grapheme cluster-like boxes based on connected components.
    """
    objects = binary_objects(binary)
    bysize = sorted(objects, key=sl.area)
    boxmap = np.zeros(binary.shape, dtype)
    for o in bysize:
        if sl.area(o)**.5 < threshold[0]*scale:
            continue
        if sl.area(o)**.5 > threshold[1]*scale:
            continue
        boxmap[o] = 1
    return boxmap


def compute_lines(segmentation: np.ndarray, scale: float) -> List[record]:
    """Given a line segmentation map, computes a list
    of tuples consisting of 2D slices and masked images."""
    logger.debug('Convert segmentation to lines')
    lobjects = morph.find_objects(segmentation)
    lines = []
    for i, o in enumerate(lobjects):
        if o is None:
            continue
        if sl.dim1(o) < 2*scale or sl.dim0(o) < scale:
            continue
        mask = (segmentation[o] == i+1)
        if np.amax(mask) == 0:
            continue
        result = record()
        result.label = i+1
        result.bounds = o
        result.mask = mask
        lines.append(result)
    return lines


def compute_separators_morph(binary: np.ndarray, scale: float, sepwiden: int = 10, maxcolseps: int = 2) -> np.ndarray:
    """Finds vertical black lines corresponding to column separators."""
    logger.debug('Finding vertical black column lines')
    d0 = int(max(5, scale/4))
    d1 = int(max(5, scale)) + sepwiden
    thick = morph.r_dilation(binary, (d0, d1))
    vert = morph.rb_opening(thick, (10*scale, 1))
    vert = morph.r_erosion(vert, (d0//2, sepwiden))
    vert = morph.select_regions(vert, sl.dim1, min=3, nbest=2*maxcolseps)
    vert = morph.select_regions(vert, sl.dim0, min=20*scale, nbest=maxcolseps)
    return vert


def compute_colseps_conv(binary: np.ndarray, scale: float = 1.0,
                         minheight: int = 10, maxcolseps: int = 2) -> np.ndarray:
    """
    Find column separators by convolution and thresholding.

    Args:
        binary:
        scale:
        minheight:
        maxcolseps:

    Returns:
        Separators
    """
    logger.debug(f'Finding max {maxcolseps} column separators')
    # find vertical whitespace by thresholding
    smoothed = gaussian_filter(1.0*binary, (scale, scale*0.5))
    smoothed = uniform_filter(smoothed, (5.0*scale, 1))
    thresh = (smoothed < np.amax(smoothed)*0.1)
    # find column edges by filtering
    grad = gaussian_filter(1.0*binary, (scale, scale*0.5), order=(0, 1))
    grad = uniform_filter(grad, (10.0*scale, 1))
    grad = (grad > 0.5*np.amax(grad))
    # combine edges and whitespace
    seps = np.minimum(thresh, maximum_filter(grad, (int(scale), int(5*scale))))
    seps = maximum_filter(seps, (int(2*scale), 1))
    # select only the biggest column separators
    seps = morph.select_regions(seps, sl.dim0, min=minheight*scale,
                                nbest=maxcolseps)
    return seps


def compute_black_colseps(binary: np.ndarray, scale: float, maxcolseps: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes column separators from vertical black lines.

    Args:
        binary: Numpy array of the binary image
        scale:
        maxcolseps:

    Returns:
        (colseps, binary):
    """
    logger.debug('Extract vertical black column separators from lines')
    seps = compute_separators_morph(binary, scale, maxcolseps)
    colseps = np.maximum(compute_colseps_conv(binary, scale, maxcolseps=maxcolseps), seps)
    binary = np.minimum(binary, 1-seps)
    return colseps, binary


def compute_white_colseps(binary: np.ndarray, scale: float, maxcolseps: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes column separators either from vertical black lines or whitespace.

    Args:
        binary: Numpy array of the binary image
        scale:

    Returns:
        colseps:
    """
    return compute_colseps_conv(binary, scale, maxcolseps=maxcolseps)


def norm_max(v: np.ndarray) -> np.ndarray:
    """
    Normalizes the input array by maximum value.
    """
    return v/np.amax(v)


def compute_gradmaps(binary: np.ndarray, scale: float, gauss: bool = False):
    """
    Use gradient filtering to find baselines

    Args:
        binary:
        scale:
        gauss: Use gaussian instead of uniform filtering

    Returns:
        (bottom, top, boxmap)
    """
    # use gradient filtering to find baselines
    logger.debug('Computing gradient maps')
    boxmap = compute_boxmap(binary, scale)
    cleaned = boxmap*binary
    if gauss:
        grad = gaussian_filter(1.0*cleaned, (0.3*scale, 6*scale), order=(1, 0))
    else:
        grad = gaussian_filter(1.0*cleaned, (max(4, 0.3*scale),
                                             scale), order=(1, 0))
        grad = uniform_filter(grad, (1, 6*scale))
    bottom = norm_max((grad < 0)*(-grad))
    top = norm_max((grad > 0)*grad)
    return bottom, top, boxmap


def compute_line_seeds(binary: np.ndarray, bottom: np.ndarray, top: np.ndarray,
                       colseps: np.ndarray, scale: float, threshold: float = 0.2) -> np.ndarray:
    """
    Base on gradient maps, computes candidates for baselines and xheights.
    Then, it marks the regions between the two as a line seed.
    """
    logger.debug('Finding line seeds')
    vrange = int(scale)
    bmarked = maximum_filter(bottom == maximum_filter(bottom, (vrange, 0)),
                             (2, 2))
    bmarked = bmarked * (bottom > threshold*np.amax(bottom)*threshold)*(1-colseps)
    tmarked = maximum_filter(top == maximum_filter(top, (vrange, 0)), (2, 2))
    tmarked = tmarked * (top > threshold*np.amax(top)*threshold/2)*(1-colseps)
    tmarked = maximum_filter(tmarked, (1, 20))
    seeds = np.zeros(binary.shape, 'i')
    delta = max(3, int(scale/2))
    for x in range(bmarked.shape[1]):
        transitions = sorted([(y, 1) for y in find(bmarked[:, x])] +
                             [(y, 0) for y in find(tmarked[:, x])])[::-1]
        transitions += [(0, 0)]
        for ls in range(len(transitions)-1):
            y0, s0 = transitions[ls]
            if s0 == 0:
                continue
            seeds[y0-delta:y0, x] = 1
            y1, s1 = transitions[ls+1]
            if s1 == 0 and (y0-y1) < 5*scale:
                seeds[y1:y0, x] = 1
    seeds = maximum_filter(seeds, (1, int(1+scale)))
    seeds = seeds * (1-colseps)
    seeds, _ = morph.label(seeds)
    return seeds


def remove_hlines(binary: np.ndarray, scale: float, maxsize: int = 10) -> np.ndarray:
    """
    Removes horizontal black lines that only interfere with page segmentation.

        Args:
            binary:
            scale:
            maxsize: maximum size of removed lines

        Returns:
            numpy.ndarray containing the filtered image.

    """
    logger.debug('Filtering horizontal lines')
    labels, _ = morph.label(binary)
    objects = morph.find_objects(labels)
    for i, b in enumerate(objects):
        if sl.width(b) > maxsize*scale:
            labels[b][labels[b] == i+1] = 0
    return np.array(labels != 0, 'B')


def rotate_lines(lines: np.ndarray, angle: float, offset: int) -> np.ndarray:
    """
    Rotates line bounding boxes around the origin and adding and offset.
    """
    logger.debug(f'Rotate line coordinates by {angle} with offset {offset}')
    angle = np.radians(angle)
    r = np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
    p = np.array(lines).reshape((-1, 2))
    offset = np.array([2*offset])
    p = p.dot(r).reshape((-1, 4)).astype(int) + offset
    x = np.sort(p[:, [0, 2]])
    y = np.sort(p[:, [1, 3]])
    return np.column_stack((x.flatten(), y.flatten())).reshape(-1, 4)


def segment(im: PIL.Image.Image,
            text_direction: str = 'horizontal-lr',
            scale: Optional[float] = None,
            maxcolseps: float = 2,
            black_colseps: bool = False,
            no_hlines: bool = True,
            pad: Union[int, Tuple[int, int]] = 0,
            mask: Optional[np.ndarray] = None,
            reading_order_fn: Callable = reading_order) -> Segmentation:
    """
    Segments a page into text lines.

    Segments a page into text lines and returns the absolute coordinates of
    each line in reading order.

    Args:
        im: A bi-level page of mode '1' or 'L'
        text_direction: Principal direction of the text
                        (horizontal-lr/rl/vertical-lr/rl)
        scale: Scale of the image. Will be auto-determined if set to `None`.
        maxcolseps: Maximum number of whitespace column separators
        black_colseps: Whether column separators are assumed to be vertical
                       black lines or not
        no_hlines: Switch for small horizontal line removal.
        pad: Padding to add to line bounding boxes. If int the same padding is
             used both left and right. If a 2-tuple, uses (padding_left,
             padding_right).
        mask: A bi-level mask image of the same size as `im` where 0-valued
              regions are ignored for segmentation purposes. Disables column
              detection.
        reading_order_fn: Function to call to order line output. Callable
                          accepting a list of slices (y, x) and a text
                          direction in (`rl`, `lr`).

    Returns:
        A :class:`kraken.containers.Segmentation` class containing reading
        order sorted bounding box-type lines as
        :class:`kraken.containers.BBoxLine` records.

    Raises:
        KrakenInputException: if the input image is not binarized or the text
                              direction is invalid.
    """
    im_str = get_im_str(im)
    logger.info(f'Segmenting {im_str}')

    if im.mode != '1' and not is_bitonal(im):
        logger.error(f'Image {im_str} is not bi-level')
        raise KrakenInputException(f'Image {im_str} is not bi-level')

    # rotate input image for vertical lines
    if text_direction.startswith('horizontal'):
        angle = 0
        offset = (0, 0)
    elif text_direction == 'vertical-lr':
        angle = 270
        offset = (0, im.size[1])
    elif text_direction == 'vertical-rl':
        angle = 90
        offset = (im.size[0], 0)
    else:
        logger.error(f'Invalid text direction \'{text_direction}\'')
        raise KrakenInputException(f'Invalid text direction {text_direction}')

    logger.debug(f'Rotating input image by {angle} degrees')
    im = im.rotate(angle, expand=True)

    a = pil2array(im)
    binary = np.array(a > 0.5*(np.amin(a) + np.amax(a)), 'i')
    binary = 1 - binary

    _, ccs = morph.label(1 - binary)
    if ccs > np.dot(*im.size)/(30*30):
        logger.warning(f'Too many connected components for a page image: {ccs}')
        return {'text_direction': text_direction, 'boxes':  []}

    if not scale:
        scale = estimate_scale(binary)

    if no_hlines:
        binary = remove_hlines(binary, scale)
    # emptyish images will cause exceptions here.

    try:
        if mask:
            if mask.mode != '1' and not is_bitonal(mask):
                logger.error('Mask is not bitonal')
                raise KrakenInputException('Mask is not bitonal')
            mask = mask.convert('1')
            if mask.size != im.size:
                logger.error(f'Mask size {mask.size} doesn\'t match image size {im.size}')
                raise KrakenInputException(f'Mask size {mask.size} doesn\'t match image size {im.size}')
            logger.info('Masking enabled in segmenter. Disabling column detection.')
            mask = mask.rotate(angle, expand=True)
            colseps = pil2array(mask)
        elif black_colseps:
            colseps, binary = compute_black_colseps(binary, scale, maxcolseps)
        else:
            colseps = compute_white_colseps(binary, scale, maxcolseps)
    except ValueError:
        logger.warning(f'Exception in column finder (probably empty image) for {im_str}')
        return {'text_direction': text_direction, 'boxes':  []}

    bottom, top, boxmap = compute_gradmaps(binary, scale)
    seeds = compute_line_seeds(binary, bottom, top, colseps, scale)
    llabels = morph.propagate_labels(boxmap, seeds, conflict=0)
    spread = morph.spread_labels(seeds, maxdist=scale)
    llabels = np.where(llabels > 0, llabels, spread*binary)
    segmentation = llabels*binary

    lines = compute_lines(segmentation, scale)
    order = reading_order_fn([line.bounds for line in lines], text_direction[-2:])
    lsort = topsort(order)
    lines = [lines[i].bounds for i in lsort]
    lines = [(s2.start, s1.start, s2.stop, s1.stop) for s1, s2 in lines]

    if isinstance(pad, int):
        pad = (pad, pad)
    lines = [(max(x[0]-pad[0], 0), x[1], min(x[2]+pad[1], im.size[0]), x[3]) for x in lines]
    lines = [BBoxLine(id=str(uuid.uuid4()), bbox=line) for line in rotate_lines(lines, 360-angle, offset).tolist()]

    return Segmentation(text_direction=text_direction,
                        imagename=getattr(im, 'filename', None),
                        type='bbox',
                        regions=None,
                        line_orders=None,
                        lines=lines,
                        script_detection=False)



#
# Copyright 2019 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.blla
~~~~~~~~~~~

Trainable layout analysis tools for kraken for line and region detection. The
line recognizer uses the baseline paradigm.
"""

import logging
import uuid
from typing import Any, Callable, Dict, List, Literal, Optional, Union

import importlib_resources
import numpy as np
import PIL
import shapely.geometry as geom
import torch
import torch.nn.functional as F
import torchvision.transforms as tf
from scipy.ndimage import gaussian_filter
from skimage.filters import sobel

from kraken.containers import BaselineLine, Region, Segmentation
from kraken.lib import dataset, vgsl
from kraken.lib.exceptions import (KrakenInputException,
                                   KrakenInvalidModelException)
from kraken.lib.segmentation import (calculate_polygonal_environment,
                                     is_in_region, neural_reading_order,
                                     polygonal_reading_order,
                                     scale_polygonal_lines, scale_regions,
                                     vectorize_lines, vectorize_regions)
from kraken.lib.util import get_im_str, is_bitonal

__all__ = ['segment']

logger = logging.getLogger(__name__)


def compute_segmentation_map(im: PIL.Image.Image,
                             mask: Optional[np.ndarray] = None,
                             model: vgsl.TorchVGSLModel = None,
                             device: str = 'cpu',
                             autocast: bool = False) -> Dict[str, Any]:
    """
    Args:
        im: Input image
        mask: A bi-level mask array of the same size as `im` where 0-valued
              regions are ignored for segmentation purposes. Disables column
              detection.
        model: A TorchVGSLModel containing a segmentation model.
        device: The target device to run the neural network on.
        autocast: Runs the model with automatic mixed precision

    Returns:
        A dictionary containing the heatmaps ('heatmap', torch.Tensor), class
        map ('cls_map', Dict[str, Dict[str, int]]), the bounding regions for
        polygonization purposes ('bounding_regions', List[str]), the scale
        between the input image and the network output ('scale', float), and
        the scaled input image to the network ('scal_im', PIL.Image.Image).

    Raises:
        KrakenInputException: When given an invalid mask.
    """

    if model.input[1] == 1 and model.one_channel_mode == '1' and not is_bitonal(im):
        logger.warning('Running binary model on non-binary input image '
                       '(mode {}). This will result in severely degraded '
                       'performance'.format(im.mode))

    model.eval()
    model.to(device)

    batch, channels, height, width = model.input
    padding = model.user_metadata['hyper_params']['padding'] if 'padding' in model.user_metadata['hyper_params'] else (0, 0)
    # expand padding to 4-tuple (left, right, top, bottom)
    if isinstance(padding, int):
        padding = (padding,) * 4
    elif len(padding) == 2:
        padding = (padding[0], padding[0], padding[1], padding[1])

    transforms = dataset.ImageInputTransforms(batch, height, width, channels, padding, valid_norm=False)
    tf_idx, _ = next(filter(lambda x: isinstance(x[1], tf.ToTensor), enumerate(transforms.transforms)))
    res_tf = tf.Compose(transforms.transforms[:tf_idx])
    scal_im = np.array(res_tf(im).convert('L'))

    tensor_im = transforms(im)
    if mask:
        if mask.mode != '1' and not is_bitonal(mask):
            logger.error('Mask is not bitonal')
            raise KrakenInputException('Mask is not bitonal')
        mask = mask.convert('1')
        if mask.size != im.size:
            logger.error('Mask size {mask.size} doesn\'t match image size {im.size}')
            raise KrakenInputException('Mask size {mask.size} doesn\'t match image size {im.size}')
        logger.info('Masking enabled in segmenter.')
        tensor_im[~transforms(mask).bool()] = 0

    with torch.autocast(device_type=device.split(":")[0], enabled=autocast):
        with torch.no_grad():
            logger.debug('Running network forward pass')
            o, _ = model.nn(tensor_im.unsqueeze(0).to(device))

    logger.debug('Upsampling network output')
    o = F.interpolate(o, size=scal_im.shape)
    # remove padding
    padding = [pad if pad else None for pad in padding]
    padding[1] = -padding[1] if padding[1] else None
    padding[3] = -padding[3] if padding[3] else None
    o = o[:, :, padding[2]:padding[3], padding[0]:padding[1]]
    scal_im = scal_im[padding[2]:padding[3], padding[0]:padding[1]]

    o = o.squeeze().cpu().float().numpy()
    scale = np.divide(im.size, o.shape[:0:-1])

    bounding_regions = model.user_metadata['bounding_regions'] if 'bounding_regions' in model.user_metadata else None
    return {'heatmap': o,
            'cls_map': model.user_metadata['class_mapping'],
            'bounding_regions': bounding_regions,
            'scale': scale,
            'scal_im': scal_im}


def vec_regions(heatmap: torch.Tensor, cls_map: Dict, scale: float, **kwargs) -> Dict[str, List[Region]]:
    """
    Computes regions from a stack of heatmaps, a class mapping, and scaling
    factor.

    Args:
        heatmap: A stack of heatmaps of shape `NxHxW` output from the network.
        cls_map: Dictionary mapping string identifiers to indices on the stack
                 of heatmaps.
        scale: Scaling factor between heatmap and unscaled input image.

    Returns:
        A dictionary containing a key for each region type with a list of
        regions inside.
    """
    logger.info('Vectorizing regions')
    regions = {}
    for region_type, idx in cls_map['regions'].items():
        logger.debug(f'Vectorizing regions of type {region_type}')
        regions[region_type] = vectorize_regions(heatmap[idx])
    for reg_type, regs in regions.items():
        regions[reg_type] = [Region(id=str(uuid.uuid4()), boundary=x, tags={'type': reg_type}) for x in scale_regions(regs, scale)]
    return regions


def vec_lines(heatmap: torch.Tensor,
              cls_map: Dict[str, Dict[str, int]],
              scale: float,
              text_direction: str = 'horizontal-lr',
              regions: List[np.ndarray] = None,
              scal_im: np.ndarray = None,
              suppl_obj: List[np.ndarray] = None,
              topline: Optional[bool] = False,
              raise_on_error: bool = False,
              **kwargs) -> List[Dict[str, Any]]:
    r"""
    Computes lines from a stack of heatmaps, a class mapping, and scaling
    factor.

    Args:
        heatmap: A stack of heatmaps of shape `NxHxW` output from the network.
        cls_map: Dictionary mapping string identifiers to indices on the stack
                 of heatmaps.
        scale: Scaling factor between heatmap and unscaled input image.
        text_direction: Text directions used as hints in the reading order
                        algorithm.
        regions: Regions to be used as boundaries during polygonization and
                 atomic blocks during reading order determination for lines
                 contained within.
        scal_im: A numpy array containing the scaled input image.
        suppl_obj: Supplementary objects which are used as boundaries during
                   polygonization.
        topline: True for a topline, False for baseline, or None for a
                 centerline.
        raise_on_error: Raises error instead of logging them when they are
                        not-blocking

    Returns:
        A list of dictionaries containing the baselines, bounding polygons, and
        line type in reading order:

        .. code-block::
           :force:

            [{'script': '$baseline_type', baseline': [[x0, y0], [x1, y1], ..., [x_n, y_n]], 'boundary': [[x0, y0, x1, y1], ... [x_m, y_m]]},
             {'script': '$baseline_type', baseline': [[x0, ...]], 'boundary': [[x0, ...]]},
             {'script': '$baseline_type', baseline': [[x0, ...]], 'boundary': [[x0, ...]]},
             ...
            ]
    """

    st_sep = cls_map['aux']['_start_separator']
    end_sep = cls_map['aux']['_end_separator']

    logger.info('Vectorizing baselines')
    baselines = []
    for bl_type, idx in cls_map['baselines'].items():
        logger.debug(f'Vectorizing lines of type {bl_type}')
        baselines.extend([(bl_type, x) for x in vectorize_lines(heatmap[(st_sep, end_sep, idx), :, :], text_direction=text_direction[:-3])])
    logger.debug('Polygonizing lines')

    im_feats = gaussian_filter(sobel(scal_im), 0.5)

    lines = []
    reg_pols = [geom.Polygon(x) for x in regions]
    for bl_idx in range(len(baselines)):
        bl = baselines[bl_idx]
        bl_ls = geom.LineString(bl[1])
        suppl_obj = [x[1] for x in baselines[:bl_idx] + baselines[bl_idx+1:]]
        for reg_idx, reg_pol in enumerate(reg_pols):
            if is_in_region(bl_ls, reg_pol):
                suppl_obj.append(regions[reg_idx])
        pol = calculate_polygonal_environment(baselines=[bl[1]],
                                              im_feats=im_feats,
                                              suppl_obj=suppl_obj,
                                              topline=topline,
                                              raise_on_error=raise_on_error)
        if pol[0] is not None:
            lines.append((bl[0], bl[1], pol[0]))

    logger.debug('Scaling vectorized lines')
    sc = scale_polygonal_lines([x[1:] for x in lines], scale)

    lines = list(zip([x[0] for x in lines], [x[0] for x in sc], [x[1] for x in sc]))
    return [{'tags': {'type': bl_type}, 'baseline': bl, 'boundary': pl} for bl_type, bl, pl in lines]


def segment(im: PIL.Image.Image,
            text_direction: Literal['horizontal-lr', 'horizontal-rl', 'vertical-lr', 'vertical-rl'] = 'horizontal-lr',
            mask: Optional[np.ndarray] = None,
            reading_order_fn: Callable = polygonal_reading_order,
            model: Union[List[vgsl.TorchVGSLModel], vgsl.TorchVGSLModel] = None,
            device: str = 'cpu',
            raise_on_error: bool = False,
            autocast: bool = False) -> Segmentation:
    r"""
    Segments a page into text lines using the baseline segmenter.

    Segments a page into text lines and returns the polyline formed by each
    baseline and their estimated environment.

    Args:
        im: Input image. The mode can generally be anything but it is possible
            to supply a binarized-input-only model which requires accordingly
            treated images.
        text_direction: Passed-through value for serialization.serialize.
        mask: A bi-level mask image of the same size as `im` where 0-valued
              regions are ignored for segmentation purposes. Disables column
              detection.
        reading_order_fn: Function to determine the reading order.  Has to
                          accept a list of tuples (baselines, polygon) and a
                          text direction (`lr` or `rl`).
        model: One or more TorchVGSLModel containing a segmentation model. If
               none is given a default model will be loaded.
        device: The target device to run the neural network on.
        raise_on_error: Raises error instead of logging them when they are
                        not-blocking
        autocast: Runs the model with automatic mixed precision

    Returns:
        A :class:`kraken.containers.Segmentation` class containing reading
        order sorted baselines (polylines) and their respective polygonal
        boundaries as :class:`kraken.containers.BaselineLine` records. The
        last and first point of each boundary polygon are connected.

    Raises:
        KrakenInvalidModelException: if the given model is not a valid
                                     segmentation model.
        KrakenInputException: if the mask is not bitonal or does not match the
                              image size.

    Notes:
        Multi-model operation is most useful for combining one or more region
        detection models and one text line model. Detected lines from all
        models are simply combined without any merging or duplicate detection
        so the chance of the same line appearing multiple times in the output
        are high. In addition, neural reading order determination is disabled
        when more than one model outputs lines.
    """
    if model is None:
        logger.info('No segmentation model given. Loading default model.')
        model = vgsl.TorchVGSLModel.load_model(importlib_resources.files(__name__).joinpath('blla.mlmodel'))

    if isinstance(model, vgsl.TorchVGSLModel):
        model = [model]

    for nn in model:
        if nn.model_type != 'segmentation':
            raise KrakenInvalidModelException(f'Invalid model type {nn.model_type} for {nn}')
        if 'class_mapping' not in nn.user_metadata:
            raise KrakenInvalidModelException(f'Segmentation model {nn} does not contain valid class mapping')

    im_str = get_im_str(im)
    logger.info(f'Segmenting {im_str}')

    lines = []
    order = None
    regions = {}
    multi_lines = False
    # flag to indicate that multiple models produced line output -> disable
    # neural reading order
    for net in model:
        if 'topline' in net.user_metadata:
            loc = {None: 'center',
                   True: 'top',
                   False: 'bottom'}[net.user_metadata['topline']]
            logger.debug(f'Baseline location: {loc}')
        rets = compute_segmentation_map(im, mask, net, device, autocast=autocast)
        _regions = vec_regions(**rets)
        for reg_key, reg_val in vec_regions(**rets).items():
            if reg_key not in regions:
                regions[reg_key] = []
            regions[reg_key].extend(reg_val)

        # flatten regions for line ordering/fetch bounding regions
        line_regs = []
        suppl_obj = []
        for cls, regs in _regions.items():
            line_regs.extend(regs)
            if rets['bounding_regions'] is not None and cls in rets['bounding_regions']:
                suppl_obj.extend(regs)
        # convert back to net scale
        suppl_obj = scale_regions([x.boundary for x in suppl_obj], 1/rets['scale'])
        line_regs = scale_regions([x.boundary for x in line_regs], 1/rets['scale'])

        _lines = vec_lines(**rets,
                           regions=line_regs,
                           text_direction=text_direction,
                           suppl_obj=suppl_obj,
                           topline=net.user_metadata['topline'] if 'topline' in net.user_metadata else False,
                           raise_on_error=raise_on_error)

        if 'ro_model' in net.aux_layers:
            logger.info(f'Using reading order model found in segmentation model {net}.')
            _order = neural_reading_order(lines=_lines,
                                          regions=_regions,
                                          text_direction=text_direction[-2:],
                                          model=net.aux_layers['ro_model'],
                                          im_size=im.size,
                                          class_mapping=net.user_metadata['ro_class_mapping'])
        else:
            _order = None

        if _lines and lines or multi_lines:
            multi_lines = True
            order = None
            logger.warning('Multiple models produced line output. This is '
                           'likely unintended. Suppressing neural reading '
                           'order.')
        else:
            order = _order

        lines.extend(_lines)

    if len(rets['cls_map']['baselines']) > 1:
        script_detection = True
    else:
        script_detection = False

    # create objects and assign IDs
    blls = []
    _shp_regs = {}
    for reg_type, rgs in regions.items():
        for reg in rgs:
            _shp_regs[reg.id] = geom.Polygon(reg.boundary)

    # reorder lines
    logger.debug(f'Reordering baselines with main RO function {reading_order_fn}.')
    basic_lo = reading_order_fn(lines=lines, regions=_shp_regs.values(), text_direction=text_direction[-2:])
    lines = [lines[idx] for idx in basic_lo]

    for line in lines:
        line_regs = []
        for reg_id, reg in _shp_regs.items():
            line_ls = geom.LineString(line['baseline'])
            if is_in_region(line_ls, reg):
                line_regs.append(reg_id)
        blls.append(BaselineLine(id=str(uuid.uuid4()), baseline=line['baseline'], boundary=line['boundary'], tags=line['tags'], regions=line_regs))

    return Segmentation(text_direction=text_direction,
                        imagename=getattr(im, 'filename', None),
                        type='baselines',
                        lines=blls,
                        regions=regions,
                        script_detection=script_detection,
                        line_orders=[order] if order else [])



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.ro
~~~~~~~~~~~~~~~

Command line driver for reading order training, evaluation, and handling.
"""
import logging
import pathlib

import click
from PIL import Image

from kraken.ketos.util import (_expand_gt, _validate_manifests, message,
                               to_ptl_device)
from kraken.lib.default_specs import READING_ORDER_HYPER_PARAMS

logging.captureWarnings(True)
logger = logging.getLogger('kraken')

# raise default max image size to 20k * 20k pixels
Image.MAX_IMAGE_PIXELS = 20000 ** 2


@click.command('rotrain')
@click.pass_context
@click.option('-B', '--batch-size', show_default=True, type=click.INT,
              default=READING_ORDER_HYPER_PARAMS['batch_size'], help='batch sample size')
@click.option('-o', '--output', show_default=True, type=click.Path(), default='model', help='Output model file')
@click.option('-i', '--load', show_default=True, type=click.Path(exists=True,
              readable=True), help='Load existing file to continue training')
@click.option('-F', '--freq', show_default=True, default=READING_ORDER_HYPER_PARAMS['freq'], type=click.FLOAT,
              help='Model saving and report generation frequency in epochs '
                   'during training. If frequency is >1 it must be an integer, '
                   'i.e. running validation every n-th epoch.')
@click.option('-q',
              '--quit',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['quit'],
              type=click.Choice(['early',
                                 'fixed']),
              help='Stop condition for training. Set to `early` for early stopping or `fixed` for fixed number of epochs')
@click.option('-N',
              '--epochs',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['epochs'],
              help='Number of epochs to train for')
@click.option('--min-epochs',
              show_default=True,
              default=['min_epochs'],
              help='Minimal number of epochs to train for when using early stopping.')
@click.option('--lag',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['lag'],
              help='Number of evaluations (--report frequency) to wait before stopping training without improvement')
@click.option('--min-delta',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['min_delta'],
              type=click.FLOAT,
              help='Minimum improvement between epochs to reset early stopping. By default it scales the delta by the best loss')
@click.option('-d', '--device', show_default=True, default='cpu', help='Select device to use (cpu, cuda:0, cuda:1, ...)')
@click.option('--precision', default='32', type=click.Choice(['32', '16']), help='set tensor precision')
@click.option('--optimizer',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['optimizer'],
              type=click.Choice(['Adam',
                                 'SGD',
                                 'RMSprop',
                                 'Lamb']),
              help='Select optimizer')
@click.option('-r', '--lrate', show_default=True, default=READING_ORDER_HYPER_PARAMS['lrate'], help='Learning rate')
@click.option('-m', '--momentum', show_default=True, default=READING_ORDER_HYPER_PARAMS['momentum'], help='Momentum')
@click.option('-w', '--weight-decay', show_default=True,
              default=READING_ORDER_HYPER_PARAMS['weight_decay'], help='Weight decay')
@click.option('--warmup', show_default=True, type=float,
              default=READING_ORDER_HYPER_PARAMS['warmup'], help='Number of samples to ramp up to `lrate` initial learning rate.')
@click.option('--schedule',
              show_default=True,
              type=click.Choice(['constant',
                                 '1cycle',
                                 'exponential',
                                 'cosine',
                                 'step',
                                 'reduceonplateau']),
              default=READING_ORDER_HYPER_PARAMS['schedule'],
              help='Set learning rate scheduler. For 1cycle, cycle length is determined by the `--step-size` option.')
@click.option('-g',
              '--gamma',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['gamma'],
              help='Decay factor for exponential, step, and reduceonplateau learning rate schedules')
@click.option('-ss',
              '--step-size',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['step_size'],
              help='Number of validation runs between learning rate decay for exponential and step LR schedules')
@click.option('--sched-patience',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['rop_patience'],
              help='Minimal number of validation runs between LR reduction for reduceonplateau LR schedule.')
@click.option('--cos-max',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['cos_t_max'],
              help='Epoch of minimal learning rate for cosine LR scheduler.')
@click.option('--cos-min-lr',
              show_default=True,
              default=READING_ORDER_HYPER_PARAMS['cos_min_lr'],
              help='Minimal final learning rate for cosine LR scheduler.')
@click.option('-p', '--partition', show_default=True, default=0.9,
              help='Ground truth data partition ratio between train/validation set')
@click.option('-t', '--training-files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with additional paths to training data')
@click.option('-e', '--evaluation-files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with paths to evaluation data. Overrides the `-p` parameter')
@click.option('--workers', show_default=True, default=1, type=click.IntRange(0), help='Number of worker proesses.')
@click.option('--threads', show_default=True, default=1, type=click.IntRange(1), help='Maximum size of OpenMP/BLAS thread pool.')
@click.option('--load-hyper-parameters/--no-load-hyper-parameters', show_default=True, default=False,
              help='When loading an existing model, retrieve hyper-parameters from the model')
@click.option('-f', '--format-type', type=click.Choice(['xml', 'alto', 'page']), default='xml',
              help='Sets the training data format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both baselines and a '
              'link to source images.')
@click.option('--logger', 'pl_logger', show_default=True, type=click.Choice(['tensorboard']), default=None,
              help='Logger used by PyTorch Lightning to track metrics such as loss and accuracy.')
@click.option('--log-dir', show_default=True, type=click.Path(exists=True, dir_okay=True, writable=True),
              help='Path to directory where the logger will store the logs. If not set, a directory will be created in the current working directory.')
@click.option('--level', show_default=True, type=click.Choice(['baselines', 'regions']), default='baselines',
              help='Selects level to train reading order model on.')
@click.option('--reading-order', show_default=True, default=None,
              help='Select reading order to train. Defaults to `line_implicit`/`region_implicit`')
@click.argument('ground_truth', nargs=-1, callback=_expand_gt, type=click.Path(exists=False, dir_okay=False))
def rotrain(ctx, batch_size, output, load, freq, quit, epochs, min_epochs, lag,
            min_delta, device, precision, optimizer, lrate, momentum,
            weight_decay, warmup, schedule, gamma, step_size, sched_patience,
            cos_max, cos_min_lr, partition, training_files, evaluation_files,
            workers, threads, load_hyper_parameters, format_type, pl_logger,
            log_dir, level, reading_order, ground_truth):
    """
    Trains a baseline labeling model for layout analysis
    """
    import shutil

    from threadpoolctl import threadpool_limits

    from kraken.lib.ro import ROModel
    from kraken.lib.train import KrakenTrainer

    if not (0 <= freq <= 1) and freq % 1.0 != 0:
        raise click.BadOptionUsage('freq', 'freq needs to be either in the interval [0,1.0] or a positive integer.')

    if pl_logger == 'tensorboard':
        try:
            import tensorboard  # NOQA
        except ImportError:
            raise click.BadOptionUsage('logger', 'tensorboard logger needs the `tensorboard` package installed.')

    if log_dir is None:
        log_dir = pathlib.Path.cwd()

    logger.info('Building ground truth set from {} document images'.format(len(ground_truth) + len(training_files)))

    # populate hyperparameters from command line args
    hyper_params = READING_ORDER_HYPER_PARAMS.copy()
    hyper_params.update({'batch_size': batch_size,
                         'freq': freq,
                         'quit': quit,
                         'epochs': epochs,
                         'min_epochs': min_epochs,
                         'lag': lag,
                         'min_delta': min_delta,
                         'optimizer': optimizer,
                         'lrate': lrate,
                         'momentum': momentum,
                         'weight_decay': weight_decay,
                         'warmup': warmup,
                         'schedule': schedule,
                         'gamma': gamma,
                         'step_size': step_size,
                         'rop_patience': sched_patience,
                         'cos_t_max': cos_max,
                         'cos_min_lr': cos_min_lr,
                         'pl_logger': pl_logger,
                         }
                        )

    # disable automatic partition when given evaluation set explicitly
    if evaluation_files:
        partition = 1
    ground_truth = list(ground_truth)

    # merge training_files into ground_truth list
    if training_files:
        ground_truth.extend(training_files)

    if len(ground_truth) == 0:
        raise click.UsageError('No training data was provided to the train command. Use `-t` or the `ground_truth` argument.')

    try:
        accelerator, device = to_ptl_device(device)
    except Exception as e:
        raise click.BadOptionUsage('device', str(e))

    if hyper_params['freq'] > 1:
        val_check_interval = {'check_val_every_n_epoch': int(hyper_params['freq'])}
    else:
        val_check_interval = {'val_check_interval': hyper_params['freq']}

    if load:
        model = ROModel.load_from_checkpoint(load,
                                             training_data=ground_truth,
                                             evaluation_data=evaluation_files,
                                             partition=partition,
                                             num_workers=workers,
                                             load_hyper_parameters=load_hyper_parameters,
                                             format_type=format_type)
    else:
        model = ROModel(hyper_params,
                        output=output,
                        training_data=ground_truth,
                        evaluation_data=evaluation_files,
                        partition=partition,
                        num_workers=workers,
                        load_hyper_parameters=load_hyper_parameters,
                        format_type=format_type,
                        level=level,
                        reading_order=reading_order)

    message(f'Training RO on following {level} types:')
    for k, v in model.train_set.dataset.class_mapping.items():
        message(f'  {k}\t{v}')

    if len(model.train_set) == 0:
        raise click.UsageError('No valid training data was provided to the train command. Use `-t` or the `ground_truth` argument.')

    trainer = KrakenTrainer(accelerator=accelerator,
                            devices=device,
                            max_epochs=hyper_params['epochs'] if hyper_params['quit'] == 'fixed' else -1,
                            min_epochs=hyper_params['min_epochs'],
                            enable_progress_bar=True if not ctx.meta['verbose'] else False,
                            deterministic=ctx.meta['deterministic'],
                            precision=int(precision),
                            pl_logger=pl_logger,
                            log_dir=log_dir,
                            **val_check_interval)

    with threadpool_limits(limits=threads):
        trainer.fit(model)

    if model.best_epoch == -1:
        logger.warning('Model did not improve during training.')
        ctx.exit(1)

    if not model.current_epoch:
        logger.warning('Training aborted before end of first epoch.')
        ctx.exit(1)

    if quit == 'early':
        message(f'Moving best model {model.best_model} ({model.best_metric}) to {output}_best.mlmodel')
        logger.info(f'Moving best model {model.best_model} ({model.best_metric}) to {output}_best.mlmodel')
        shutil.copy(f'{model.best_model}', f'{output}_best.mlmodel')


@click.command('roadd')
@click.pass_context
@click.option('-o', '--output', show_default=True, type=click.Path(), default='combined_seg.mlmodel', help='Combined output model file')
@click.option('-r', '--ro-model', show_default=True, type=click.Path(exists=True, readable=True), help='Reading order model to load into segmentation model')
@click.option('-i', '--seg-model', show_default=True, type=click.Path(exists=True, readable=True), help='Segmentation model to load')
def roadd(ctx, output, ro_model, seg_model):
    """
    Combines a reading order model with a segmentation model.
    """
    from kraken.lib import vgsl
    from kraken.lib.ro import ROModel

    message(f'Adding {ro_model} reading order model to {seg_model}.')
    ro_net = ROModel.load_from_checkpoint(ro_model)
    message('Line classes known to RO model:')
    for k, v in ro_net.class_mapping.items():
        message(f'  {k}\t{v}')
    seg_net = vgsl.TorchVGSLModel.load_model(seg_model)
    if seg_net.model_type != 'segmentation':
        raise click.UsageError(f'Model {seg_model} is invalid {seg_net.model_type} model (expected `segmentation`).')
    message('Line classes known to segmentation model:')
    for k, v in seg_net.user_metadata['class_mapping']['baselines'].items():
        message(f'  {k}\t{v}')
    if ro_net.class_mapping.keys() != seg_net.user_metadata['class_mapping']['baselines'].keys():
        raise click.UsageError(f'Model {seg_model} and {ro_model} class mappings mismatch.')

    seg_net.aux_layers = {'ro_model': ro_net.ro_net}
    seg_net.user_metadata['ro_class_mapping'] = ro_net.class_mapping
    message(f'Saving combined model to {output}')
    seg_net.save_model(output)



#!/usr/bin/env python
"""
Draws a transparent overlay of the forced alignment output over the input
image.
"""
import os
import re
import unicodedata
from itertools import cycle
from unicodedata import normalize

import click
from lxml import etree

cmap = cycle([(230, 25, 75, 127),
              (60, 180, 75, 127),
              (255, 225, 25, 127),
              (0, 130, 200, 127),
              (245, 130, 48, 127),
              (145, 30, 180, 127),
              (70, 240, 240, 127)])


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    value = unicodedata.normalize('NFKD', value)
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    value = re.sub(r'[-\s]+', '-', value)
    return value


def _repl_alto(fname, cuts):
    with open(fname, 'rb') as fp:
        doc = etree.parse(fp)
        lines = doc.findall('.//{*}TextLine')
        char_idx = 0
        for line, line_cuts in zip(lines, cuts.lines):
            idx = 0
            for el in line:
                if el.tag.endswith('Shape'):
                    continue
                elif el.tag.endswith('SP'):
                    idx += 1
                elif el.tag.endswith('String'):
                    str_len = len(el.get('CONTENT'))
                    # clear out all
                    for chld in el:
                        if chld.tag.endswith('Glyph'):
                            el.remove(chld)
                    for char in zip(line_cuts.prediction[idx:str_len],
                                    line_cuts.cuts[idx:str_len],
                                    line_cuts.confidences[idx:str_len]):
                        glyph = etree.SubElement(el, 'Glyph')
                        glyph.set('ID', f'char_{char_idx}')
                        char_idx += 1
                        glyph.set('CONTENT', char[0])
                        glyph.set('GC', str(char[2]))
                        pol = etree.SubElement(etree.SubElement(glyph, 'Shape'), 'Polygon')
                        pol.set('POINTS', ' '.join([str(coord) for pt in char[1] for coord in pt]))
                    idx += str_len
        with open(f'{os.path.basename(fname)}_algn.xml', 'wb') as fp:
            doc.write(fp, encoding='UTF-8', xml_declaration=True)


def _repl_page(fname, cuts):
    with open(fname, 'rb') as fp:
        doc = etree.parse(fp)
        lines = doc.findall('.//{*}TextLine')
        for line, line_cuts in zip(lines, cuts.lines):
            glyphs = line.findall('../{*}Glyph/{*}Coords')
            for glyph, cut in zip(glyphs, line_cuts):
                glyph.attrib['points'] = ' '.join([','.join([str(x) for x in pt]) for pt in cut])
        with open(f'{os.path.basename(fname)}_algn.xml', 'wb') as fp:
            doc.write(fp, encoding='UTF-8', xml_declaration=True)


@click.command()
@click.option('-f', '--format-type', type=click.Choice(['alto', 'page']), default='page',
              help='Sets the input document format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both baselines, polygons, and a '
              'link to source images.')
@click.option('-i', '--model', default=None, show_default=True, type=click.Path(exists=True),
              help='Transcription model to use.')
@click.option('-u', '--normalization', show_default=True, type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']),
              default=None,
              help='Ground truth normalization')
@click.option('-o', '--output', type=click.Choice(['xml', 'overlay']),
              show_default=True, default='overlay', help='Output mode. Either page or '
              'alto for xml output, overlay for image overlays.')
@click.argument('files', nargs=-1)
def cli(format_type, model, normalization, output, files):
    """
    A script producing overlays of lines and regions from either ALTO or
    PageXML files or run a model to do the same.
    """
    if len(files) == 0:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()

    from PIL import Image, ImageDraw

    from kraken import align
    from kraken.lib import models
    from kraken.lib.xml import XMLPage

    if format_type == 'alto':
        repl_fn = _repl_alto
    else:
        repl_fn = _repl_page

    click.echo(f'Loading model {model}')
    net = models.load_any(model)

    for doc in files:
        click.echo(f'Processing {doc} ', nl=False)
        data = XMLPage(doc)
        im = Image.open(data.imagename).convert('RGBA')
        result = align.forced_align(data.to_container(), net)
        if normalization:
            for line in data._lines:
                line["text"] = normalize(normalization, line["text"])
        im = Image.open(data.imagename).convert('RGBA')
        result = align.forced_align(data.to_container(), net)
        if output == 'overlay':
            tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
            draw = ImageDraw.Draw(tmp)
            for record in result.lines:
                for pol in record.cuts:
                    c = next(cmap)
                    draw.polygon([tuple(x) for x in pol], fill=c, outline=c[:3])
            base_image = Image.alpha_composite(im, tmp)
            base_image.save(f'high_{os.path.basename(doc)}_algn.png')
        else:
            repl_fn(doc, result)
        click.secho('\u2713', fg='green')


if __name__ == '__main__':
    cli()



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Utility functions for ground truth transcription.
"""
import base64
import logging
import uuid
from io import BytesIO
from typing import Any, Dict, List

from jinja2 import Environment, PackageLoader

from kraken.lib.exceptions import KrakenInputException
from kraken.lib.util import get_im_str

logger = logging.getLogger()


class TranscriptionInterface(object):

    def __init__(self, font=None, font_style=None):
        logging.info('Initializing transcription object.')
        logger.debug('Initializing jinja environment.')
        env = Environment(loader=PackageLoader('kraken', 'templates'), autoescape=True)
        logger.debug('Loading transcription template.')
        self.tmpl = env.get_template('layout.html')
        self.pages: List[Dict[Any, Any]] = []
        self.font = {'font': font, 'style': font_style}
        self.text_direction = 'horizontal-tb'
        self.page_idx = 1
        self.line_idx = 1
        self.seg_idx = 1

    def add_page(self, im, segmentation=None, records=None):
        """
        Adds an image to the transcription interface, optionally filling in
        information from a list of ocr_record objects.

        Args:
            im (PIL.Image): Input image
            segmentation (dict): Output of the segment method.
            records (list): A list of ocr_record objects.
        """
        im_str = get_im_str(im)
        logger.info('Adding page {} with {} lines'.format(im_str, len(segmentation) if segmentation else len(records)))
        page = {}
        fd = BytesIO()
        im.save(fd, format='png', optimize=True)
        page['index'] = self.page_idx
        self.page_idx += 1
        logger.debug('Base64 encoding image')
        page['img'] = 'data:image/png;base64,' + base64.b64encode(fd.getvalue()).decode('ascii')
        page['lines'] = []
        if records:
            logger.debug('Adding records.')
            self.text_direction = segmentation['text_direction']
            for record, bbox in zip(records, segmentation['boxes']):
                page['lines'].append({'index': self.line_idx, 'text': record.prediction,
                                      'left': 100*int(bbox[0]) / im.size[0],
                                      'top': 100*int(bbox[1]) / im.size[1],
                                      'width': 100*(bbox[2] - bbox[0])/im.size[0],
                                      'height': 100*(int(bbox[3]) - int(bbox[1]))/im.size[1],
                                      'bbox': '{}, {}, {}, {}'.format(int(bbox[0]),
                                                                      int(bbox[1]),
                                                                      int(bbox[2]),
                                                                      int(bbox[3]))})

                self.line_idx += 1
        elif segmentation:
            logger.debug('Adding segmentations.')
            self.text_direction = segmentation['text_direction']
            for bbox in segmentation['boxes']:
                page['lines'].append({'index': self.line_idx,
                                      'left': 100*int(bbox[0]) / im.size[0],
                                      'top': 100*int(bbox[1]) / im.size[1],
                                      'width': 100*(bbox[2] - bbox[0])/im.size[0],
                                      'height': 100*(int(bbox[3]) - int(bbox[1]))/im.size[1],
                                      'bbox': '{}, {}, {}, {}'.format(int(bbox[0]),
                                                                      int(bbox[1]),
                                                                      int(bbox[2]),
                                                                      int(bbox[3]))})
                self.line_idx += 1
        else:
            raise KrakenInputException('Neither segmentations nor records given')
        self.pages.append(page)

    def write(self, fd):
        """
        Writes the HTML file to a file descriptor.

        Args:
            fd (File): File descriptor (mode='rb') to write to.
        """
        logger.info('Rendering and writing transcription.')
        fd.write(self.tmpl.render(uuid=str(uuid.uuid4()), pages=self.pages,
                                  font=self.font,
                                  text_direction=self.text_direction).encode('utf-8'))



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Utility functions for data loading and training of VGSL networks.
"""
from math import factorial
from typing import TYPE_CHECKING, Dict, Literal, Optional, Sequence, Union

import numpy as np
import torch
from torch.utils.data import Dataset

from kraken.lib.exceptions import KrakenInputException
from kraken.lib.xml import XMLPage

if TYPE_CHECKING:
    from os import PathLike

__all__ = ['PairWiseROSet', 'PageWiseROSet']

import logging

logger = logging.getLogger(__name__)


class PairWiseROSet(Dataset):
    """
    Dataset for training a reading order determination model.

    Returns random pairs of lines from the same page.
    """
    def __init__(self, files: Sequence[Union['PathLike', str]] = None,
                 mode: Optional[Literal['alto', 'page', 'xml']] = 'xml',
                 level: Literal['regions', 'baselines'] = 'baselines',
                 ro_id: Optional[str] = None,
                 class_mapping: Optional[Dict[str, int]] = None):
        """
        Samples pairs lines/regions from XML files for training a reading order
        model .

        Args:
            mode: Either alto, page, xml, None. In alto, page, and xml
                  mode the baseline paths and image data is retrieved from an
                  ALTO/PageXML file. In `None` mode data is iteratively added
                  through the `add` method.
            ro_id: ID of the reading order to sample from. Defaults to
                   `line_implicit`/`region_implicit`.
        """
        super().__init__()

        self._num_pairs = 0
        self.failed_samples = []
        if class_mapping:
            self.class_mapping = class_mapping
            self.num_classes = len(class_mapping) + 1
        else:
            self.num_classes = 1
            self.class_mapping = {}

        self.data = []

        if mode in ['alto', 'page', 'xml']:
            for file in files:
                try:
                    doc = XMLPage(file, filetype=mode)
                    if level == 'baselines':
                        if not ro_id:
                            ro_id = 'line_implicit'
                        order = doc.get_sorted_lines(ro_id)
                    elif level == 'regions':
                        if not ro_id:
                            ro_id = 'region_implicit'
                        order = doc.get_sorted_regions(ro_id)
                    else:
                        raise ValueError(f'Invalid RO type {level}')
                    for el in order:
                        for tag in el.tags.values():
                            if tag not in self.class_mapping:
                                self.class_mapping[tag] = self.num_classes
                                self.num_classes += 1
                    # traverse RO and substitute features.
                    w, h = doc.image_size
                    sorted_lines = []
                    for line in order:
                        line_coords = np.array(line.baseline) / (w, h)
                        line_center = np.mean(line_coords, axis=0)
                        cl = torch.zeros(self.num_classes, dtype=torch.float)
                        # if class is not in class mapping default to None class (idx 0)
                        cl[self.class_mapping.get(line.tags['type'], 0)] = 1
                        line_data = {'type': line.tags['type'],
                                     'features': torch.cat((cl,  # one hot encoded line type
                                                            torch.tensor(line_center, dtype=torch.float),  # line center
                                                            torch.tensor(line_coords[0, :], dtype=torch.float),  # start_point coord
                                                            torch.tensor(line_coords[-1, :], dtype=torch.float),  # end point coord)
                                                            )
                                                           )
                                     }
                        sorted_lines.append(line_data)
                    if len(sorted_lines) > 1:
                        self.data.append(sorted_lines)
                        self._num_pairs += int(factorial(len(sorted_lines))/factorial(len(sorted_lines)-2))
                    else:
                        logger.info(f'Page {doc} has less than 2 lines. Skipping')
                except KrakenInputException as e:
                    logger.warning(e)
                    continue
        else:
            raise Exception('invalid dataset mode')

    def __getitem__(self, idx):
        lines = []
        while len(lines) < 2:
            lines = self.data[torch.randint(len(self.data), (1,))[0]]
        idx0, idx1 = 0, 0
        while idx0 == idx1:
            idx0, idx1 = torch.randint(len(lines), (2,))
        x = torch.cat((lines[idx0]['features'], lines[idx1]['features']))
        y = torch.tensor(0 if idx0 >= idx1 else 1, dtype=torch.float)
        return {'sample': x, 'target': y}

    def get_feature_dim(self):
        return 2 * self.num_classes + 12

    def __len__(self):
        return self._num_pairs


class PageWiseROSet(Dataset):
    """
    Dataset for training a reading order determination model.

    Returns all lines from the same page.
    """
    def __init__(self, files: Sequence[Union['PathLike', str]] = None,
                 mode: Optional[Literal['alto', 'page', 'xml']] = 'xml',
                 level: Literal['regions', 'baselines'] = 'baselines',
                 ro_id: Optional[str] = None,
                 class_mapping: Optional[Dict[str, int]] = None):
        """
        Samples pairs lines/regions from XML files for training a reading order
        model .

        Args:
            mode: Either alto, page, xml, None. In alto, page, and xml
                  mode the baseline paths and image data is retrieved from an
                  ALTO/PageXML file. In `None` mode data is iteratively added
                  through the `add` method.
            ro_id: ID of the reading order to sample from. Defaults to
                   `line_implicit`/`region_implicit`.
        """
        super().__init__()

        self.failed_samples = []
        if class_mapping:
            self.class_mapping = class_mapping
            self.num_classes = len(class_mapping) + 1
        else:
            self.num_classes = 1
            self.class_mapping = {}

        self.data = []

        if mode in ['alto', 'page', 'xml']:
            for file in files:
                try:
                    doc = XMLPage(file, filetype=mode)
                    if level == 'baselines':
                        if not ro_id:
                            ro_id = 'line_implicit'
                        order = doc.get_sorted_lines(ro_id)
                    elif level == 'regions':
                        if not ro_id:
                            ro_id = 'region_implicit'
                        order = doc.get_sorted_regions(ro_id)
                    else:
                        raise ValueError(f'Invalid RO type {level}')
                    for el in order:
                        for tag in el.tags.values():
                            if tag not in self.class_mapping:
                                self.class_mapping[tag] = self.num_classes
                                self.num_classes += 1
                    # traverse RO and substitute features.
                    w, h = doc.image_size
                    sorted_lines = []
                    for line in order:
                        line_coords = np.array(line.baseline) / (w, h)
                        line_center = np.mean(line_coords, axis=0)
                        cl = torch.zeros(self.num_classes, dtype=torch.float)
                        # if class is not in class mapping default to None class (idx 0)
                        cl[self.class_mapping.get(line.tags['type'], 0)] = 1
                        line_data = {'type': line.tags['type'],
                                     'features': torch.cat((cl,  # one hot encoded line type
                                                            torch.tensor(line_center, dtype=torch.float),  # line center
                                                            torch.tensor(line_coords[0, :], dtype=torch.float),  # start_point coord
                                                            torch.tensor(line_coords[-1, :], dtype=torch.float),  # end point coord)
                                                            )
                                                           )
                                     }
                        sorted_lines.append(line_data)
                    if len(sorted_lines) > 1:
                        self.data.append(sorted_lines)
                    else:
                        logger.info(f'Page {doc} has less than 2 lines. Skipping')
                except KrakenInputException as e:
                    logger.warning(e)
                    continue
        else:
            raise Exception('invalid dataset mode')

    def __getitem__(self, idx):
        xs = []
        ys = []
        for i in range(len(self.data[idx])):
            for j in range(len(self.data[idx])):
                if i == j and len(self.data[idx]) != 1:
                    continue
                xs.append(torch.cat((self.data[idx][i]['features'],
                                     self.data[idx][j]['features'])))
                ys.append(torch.tensor(0 if i >= j else 1, dtype=torch.float))
        return {'sample': torch.stack(xs), 'target': torch.stack(ys), 'num_lines': len(self.data[idx])}

    def get_feature_dim(self):
        return 2 * self.num_classes + 12

    def __len__(self):
        return len(self.data)



#
# Copyright 2023 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.containers
~~~~~~~~~~~~~~~~~

Container classes replacing the old dictionaries returned by kraken's
functional blocks.
"""
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple,
                    Union)

import bidi.algorithm as bd
import numpy as np

from kraken.lib.segmentation import compute_polygon_section

if TYPE_CHECKING:
    from os import PathLike


__all__ = ['BaselineLine',
           'BBoxLine',
           'Segmentation',
           'Region',
           'ocr_record',
           'BaselineOCRRecord',
           'BBoxOCRRecord',
           'ProcessingStep']


@dataclass
class ProcessingStep:
    """
    A processing step in the recognition pipeline.

    Attributes:
        id: Unique identifier
        category: Category of processing step that has been performed.
        description: Natural-language description of the process.
        settings: Dict describing the parameters of the processing step.
    """
    id: str
    category: Literal['preprocessing', 'processing', 'postprocessing']
    description: str
    settings: Dict[str, Union[Dict, str, float, int, bool]]


@dataclass
class BaselineLine:
    """
    Baseline-type line record.

    A container class for a single line in baseline + bounding polygon format,
    optionally containing a transcription, tags, or associated regions.

    Attributes:
        id: Unique identifier
        baseline: List of tuples `(x_n, y_n)` defining the baseline.
        boundary: List of tuples `(x_n, y_n)` defining the bounding polygon of
                  the line. The first and last points should be identical.
        text: Transcription of this line.
        base_dir: An optional string defining the base direction (also called
                  paragraph direction) for the BiDi algorithm. Valid values are
                  'L' or 'R'. If None is given the default auto-resolution will
                  be used.
        imagename: Path to the image associated with the line.
        tags: A dict mapping types to values.
        split: Defines whether this line is in the `train`, `validation`, or
               `test` set during training.
        regions: A list of identifiers of regions the line is associated with.
    """
    id: str
    baseline: List[Tuple[int, int]]
    boundary: List[Tuple[int, int]]
    text: Optional[str] = None
    base_dir: Optional[Literal['L', 'R']] = None
    type: str = 'baselines'
    imagename: Optional[Union[str, 'PathLike']] = None
    tags: Optional[Dict[str, str]] = None
    split: Optional[Literal['train', 'validation', 'test']] = None
    regions: Optional[List[str]] = None


@dataclass
class BBoxLine:
    """
    Bounding box-type line record.

    A container class for a single line in axis-aligned bounding box format,
    optionally containing a transcription, tags, or associated regions.

    Attributes:
        id: Unique identifier
        bbox: Tuple in form `(xmin, ymin, xmax, ymax)` defining
              the bounding box.
        text: Transcription of this line.
        base_dir: An optional string defining the base direction (also called
                  paragraph direction) for the BiDi algorithm. Valid values are
                  'L' or 'R'. If None is given the default auto-resolution will
                  be used.
        imagename: Path to the image associated with the line..
        tags: A dict mapping types to values.
        split: Defines whether this line is in the `train`, `validation`, or
               `test` set during training.
        regions: A list of identifiers of regions the line is associated with.
        text_direction: Sets the principal orientation (of the line) and
                        reading direction (of the document).
    """
    id: str
    bbox: Tuple[int, int, int, int]
    text: Optional[str] = None
    base_dir: Optional[Literal['L', 'R']] = None
    type: str = 'bbox'
    imagename: Optional[Union[str, 'PathLike']] = None
    tags: Optional[Dict[str, str]] = None
    split: Optional[Literal['train', 'validation', 'test']] = None
    regions: Optional[List[str]] = None
    text_direction: Literal['horizontal-lr', 'horizontal-rl', 'vertical-lr', 'vertical-rl'] = 'horizontal-lr'


@dataclass
class Region:
    """
    Container class of a single polygonal region.

    Attributes:
        id: Unique identifier
        boundary: List of tuples `(x_n, y_n)` defining the bounding polygon of
                  the region. The first and last points should be identical.
        imagename: Path to the image associated with the region.
        tags: A dict mapping types to values.
    """
    id: str
    boundary: List[Tuple[int, int]]
    imagename: Optional[Union[str, 'PathLike']] = None
    tags: Optional[Dict[str, str]] = None


@dataclass
class Segmentation:
    """
    A container class for segmentation or recognition results.

    In order to allow easy JSON de-/serialization, nested classes for lines
    (BaselineLine/BBoxLine) and regions (Region) are reinstantiated from their
    dictionaries.

    Attributes:
        type: Field indicating if baselines
              (:class:`kraken.containers.BaselineLine`) or bbox
              (:class:`kraken.containers.BBoxLine`) line records are in the
              segmentation.
        imagename: Path to the image associated with the segmentation.
        text_direction: Sets the principal orientation (of the line), i.e.
                        horizontal/vertical, and reading direction (of the
                        document), i.e. lr/rl.
        script_detection: Flag indicating if the line records have tags.
        lines: List of line records. Records are expected to be in a valid
               reading order.
        regions: Dict mapping types to lists of regions.
        line_orders: List of alternative reading orders for the segmentation.
                     Each reading order is a list of line indices.
    """
    type: Literal['baselines', 'bbox']
    imagename: Union[str, 'PathLike']
    text_direction: Literal['horizontal-lr', 'horizontal-rl', 'vertical-lr', 'vertical-rl']
    script_detection: bool
    lines: Optional[List[Union[BaselineLine, BBoxLine]]] = None
    regions: Optional[Dict[str, List[Region]]] = None
    line_orders: Optional[List[List[int]]] = None

    def __post_init__(self):
        if not self.regions:
            self.regions = {}
        if not self.lines:
            self.lines = []
        if not self.line_orders:
            self.line_orders = []
        if len(self.lines) and not isinstance(self.lines[0], BBoxLine) and not isinstance(self.lines[0], BaselineLine):
            line_cls = BBoxLine if self.type == 'bbox' else BaselineLine
            self.lines = [line_cls(**line) for line in self.lines]
        if len(self.regions):
            for regs in self.regions.values():
                if regs and not isinstance(regs[0], Region):
                    regs = {}
                    for k, v in self.regions.items():
                        regs[k] = [Region(**reg) for reg in v]
                    self.regions = regs
                    break


class ocr_record(ABC):
    """
    A record object containing the recognition result of a single line
    """
    base_dir = None

    def __init__(self,
                 prediction: str,
                 cuts: List[Union[Tuple[int, int], Tuple[Tuple[int, int],
                                                         Tuple[int, int],
                                                         Tuple[int, int],
                                                         Tuple[int, int]]]],
                 confidences: List[float],
                 display_order: bool = True) -> None:
        self._prediction = prediction
        self._cuts = cuts
        self._confidences = confidences
        self._display_order = display_order

    @property
    @abstractmethod
    def type(self):
        pass

    def __len__(self) -> int:
        return len(self._prediction)

    def __str__(self) -> str:
        return self._prediction

    @property
    def prediction(self) -> str:
        return self._prediction

    @property
    def cuts(self) -> List:
        return self._cuts

    @property
    def confidences(self) -> List[float]:
        return self._confidences

    def __iter__(self):
        self.idx = -1
        return self

    @abstractmethod
    def __next__(self) -> Tuple[str,
                                Union[List[Tuple[int, int]],
                                      Tuple[Tuple[int, int],
                                            Tuple[int, int],
                                            Tuple[int, int],
                                            Tuple[int, int]]],
                                float]:
        pass

    @abstractmethod
    def __getitem__(self, key: Union[int, slice]):
        pass

    @abstractmethod
    def display_order(self, base_dir) -> 'ocr_record':
        pass

    @abstractmethod
    def logical_order(self, base_dir) -> 'ocr_record':
        pass


class BaselineOCRRecord(ocr_record, BaselineLine):
    """
    A record object containing the recognition result of a single line in
    baseline format.

    Attributes:
        type: 'baselines' to indicate a baseline record
        prediction: The text predicted by the network as one continuous string.
        cuts: The absolute bounding polygons for each code point in prediction
              as a list of tuples [(x0, y0), (x1, y2), ...].
        confidences: A list of floats indicating the confidence value of each
                     code point.
        base_dir: An optional string defining the base direction (also called
                  paragraph direction) for the BiDi algorithm. Valid values are
                  'L' or 'R'. If None is given the default auto-resolution will
                  be used.
        display_order: Flag indicating the order of the code points in the
                       prediction. In display order (`True`) the n-th code
                       point in the string corresponds to the n-th leftmost
                       code point, in logical order (`False`) the n-th code
                       point corresponds to the n-th read code point. See [UAX
                       #9](https://unicode.org/reports/tr9) for more details.

    Notes:
        When slicing the record the behavior of the cuts is changed from
        earlier versions of kraken. Instead of returning per-character bounding
        polygons a single polygons section of the line bounding polygon
        starting at the first and extending to the last code point emitted by
        the network is returned. This aids numerical stability when computing
        aggregated bounding polygons such as for words. Individual code point
        bounding polygons are still accessible through the `cuts` attribute or
        by iterating over the record code point by code point.
    """
    type = 'baselines'

    def __init__(self, prediction: str,
                 cuts: List[Tuple[int, int]],
                 confidences: List[float],
                 line: Union[BaselineLine, Dict[str, Any]],
                 base_dir: Optional[Literal['L', 'R']] = None,
                 display_order: bool = True) -> None:
        if not isinstance(line, dict):
            line = asdict(line)
        if line['type'] != 'baselines':
            raise TypeError('Invalid argument type (non-baseline line)')
        BaselineLine.__init__(self, **line)
        self._line_base_dir = self.base_dir
        self.base_dir = base_dir
        ocr_record.__init__(self, prediction, cuts, confidences, display_order)

    def __repr__(self) -> str:
        return f'pred: {self.prediction} baseline: {self.baseline} boundary: {self.boundary} confidences: {self.confidences}'

    def __next__(self) -> Tuple[str, int, float]:
        if self.idx + 1 < len(self):
            self.idx += 1
            return (self.prediction[self.idx],
                    compute_polygon_section(self.baseline,
                                            self.boundary,
                                            self.cuts[self.idx][0],
                                            self.cuts[self.idx][1]),
                    self.confidences[self.idx])
        else:
            raise StopIteration

    def _get_raw_item(self, key: int):
        if key < 0:
            key += len(self)
        if key >= len(self):
            raise IndexError('Index (%d) is out of range' % key)
        return (self.prediction[key],
                self._cuts[key],
                self.confidences[key])

    def __getitem__(self, key: Union[int, slice]):
        if isinstance(key, slice):
            recs = [self._get_raw_item(i) for i in range(*key.indices(len(self)))]
            prediction = ''.join([x[0] for x in recs])
            flat_offsets = sum((tuple(x[1]) for x in recs), ())
            cut = compute_polygon_section(self.baseline,
                                          self.boundary,
                                          min(flat_offsets),
                                          max(flat_offsets))
            confidence = np.mean([x[2] for x in recs])
            return (prediction, cut, confidence)
        elif isinstance(key, int):
            pred, cut, confidence = self._get_raw_item(key)
            return (pred,
                    compute_polygon_section(self.baseline, self.boundary, cut[0], cut[1]),
                    confidence)
        else:
            raise TypeError('Invalid argument type')

    @property
    def cuts(self) -> List[Tuple[int, int]]:
        return tuple([compute_polygon_section(self.baseline, self.boundary, cut[0], cut[1]) for cut in self._cuts])

    def logical_order(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BaselineOCRRecord':
        """
        Returns the OCR record in Unicode logical order, i.e. in the order the
        characters in the line would be read by a human.

        Args:
            base_dir: An optional string defining the base direction (also
                      called paragraph direction) for the BiDi algorithm. Valid
                      values are 'L' or 'R'. If None is given the default
                      auto-resolution will be used.
        """
        if self._display_order:
            return self._reorder(base_dir)
        else:
            return self

    def display_order(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BaselineOCRRecord':
        """
        Returns the OCR record in Unicode display order, i.e. ordered from left
        to right inside the line.

        Args:
            base_dir: An optional string defining the base direction (also
                      called paragraph direction) for the BiDi algorithm. Valid
                      values are 'L' or 'R'. If None is given the default
                      auto-resolution will be used.
        """
        if self._display_order:
            return self
        else:
            return self._reorder(base_dir)

    def _reorder(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BaselineOCRRecord':
        """
        Reorder the record using the BiDi algorithm.
        """
        storage = bd.get_empty_storage()

        if base_dir not in ('L', 'R'):
            base_level = bd.get_base_level(self._prediction)
        else:
            base_level = {'L': 0, 'R': 1}[base_dir]

        storage['base_level'] = base_level
        storage['base_dir'] = ('L', 'R')[base_level]
        bd.get_embedding_levels(self._prediction, storage)
        bd.explicit_embed_and_overrides(storage)
        bd.resolve_weak_types(storage)
        bd.resolve_neutral_types(storage, False)
        bd.resolve_implicit_levels(storage, False)
        for i, j in enumerate(zip(self._prediction, self._cuts, self._confidences)):
            storage['chars'][i]['record'] = j
        bd.reorder_resolved_levels(storage, False)
        bd.apply_mirroring(storage, False)
        prediction = ''
        cuts = []
        confidences = []
        for ch in storage['chars']:
            # code point may have been mirrored
            prediction = prediction + ch['ch']
            cuts.append(ch['record'][1])
            confidences.append(ch['record'][2])
        line = BaselineLine(id=self.id,
                            baseline=self.baseline,
                            boundary=self.boundary,
                            text=self.text,
                            base_dir=self._line_base_dir,
                            imagename=self.imagename,
                            tags=self.tags,
                            split=self.split,
                            regions=self.regions)

        rec = BaselineOCRRecord(prediction=prediction,
                                cuts=cuts,
                                confidences=confidences,
                                line=line,
                                base_dir=base_dir,
                                display_order=not self._display_order)
        return rec


class BBoxOCRRecord(ocr_record, BBoxLine):
    """
    A record object containing the recognition result of a single line in
    bbox format.

    Attributes:
        type: 'bbox' to indicate a bounding box record
        prediction: The text predicted by the network as one continuous string.
        cuts: The absolute bounding polygons for each code point in prediction
              as a list of 4-tuples `((x0, y0), (x1, y0), (x1, y1), (x0, y1))`.
        confidences: A list of floats indicating the confidence value of each
                     code point.
        base_dir: An optional string defining the base direction (also called
                  paragraph direction) for the BiDi algorithm. Valid values are
                  'L' or 'R'. If None is given the default auto-resolution will
                  be used.
        display_order: Flag indicating the order of the code points in the
                       prediction. In display order (`True`) the n-th code
                       point in the string corresponds to the n-th leftmost
                       code point, in logical order (`False`) the n-th code
                       point corresponds to the n-th read code point. See [UAX
                       #9](https://unicode.org/reports/tr9) for more details.

    Notes:
        When slicing the record the behavior of the cuts is changed from
        earlier versions of kraken. Instead of returning per-character bounding
        polygons a single polygons section of the line bounding polygon
        starting at the first and extending to the last code point emitted by
        the network is returned. This aids numerical stability when computing
        aggregated bounding polygons such as for words. Individual code point
        bounding polygons are still accessible through the `cuts` attribute or
        by iterating over the record code point by code point.
    """
    type = 'bbox'

    def __init__(self,
                 prediction: str,
                 cuts: List[Tuple[Tuple[int, int],
                                  Tuple[int, int],
                                  Tuple[int, int],
                                  Tuple[int, int]]],
                 confidences: List[float],
                 line: Union[BBoxLine, Dict[str, Any]],
                 base_dir: Optional[Literal['L', 'R']] = None,
                 display_order: bool = True) -> None:
        if not isinstance(line, dict):
            line = asdict(line)
        if line['type'] != 'bbox':
            raise TypeError('Invalid argument type (non-bbox line)')
        BBoxLine.__init__(self, **line)
        self._line_base_dir = self.base_dir
        self.base_dir = base_dir
        ocr_record.__init__(self, prediction, cuts, confidences, display_order)

    def __repr__(self) -> str:
        return f'pred: {self.prediction} bbox: {self.bbox} confidences: {self.confidences}'

    def __next__(self) -> Tuple[str, int, float]:
        if self.idx + 1 < len(self):
            self.idx += 1
            return (self.prediction[self.idx],
                    self.cuts[self.idx],
                    self.confidences[self.idx])
        else:
            raise StopIteration

    def _get_raw_item(self, key: int):
        if key < 0:
            key += len(self)
        if key >= len(self):
            raise IndexError('Index (%d) is out of range' % key)
        return (self.prediction[key],
                self.cuts[key],
                self.confidences[key])

    def __getitem__(self, key: Union[int, slice]):
        if isinstance(key, slice):
            recs = [self._get_raw_item(i) for i in range(*key.indices(len(self)))]
            prediction = ''.join([x[0] for x in recs])
            box = [x[1] for x in recs]
            flat_box = [point for pol in box for point in pol]
            flat_box = [x for point in flat_box for x in point]
            min_x, max_x = min(flat_box[::2]), max(flat_box[::2])
            min_y, max_y = min(flat_box[1::2]), max(flat_box[1::2])
            cut = ((min_x, min_y), (max_x, min_y), (max_x, max_y), (min_x, max_y))
            confidence = np.mean([x[2] for x in recs])
            return (prediction, cut, confidence)
        elif isinstance(key, int):
            return self._get_raw_item(key)
        else:
            raise TypeError('Invalid argument type')

    def logical_order(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BBoxOCRRecord':
        """
        Returns the OCR record in Unicode logical order, i.e. in the order the
        characters in the line would be read by a human.

        Args:
            base_dir: An optional string defining the base direction (also
                      called paragraph direction) for the BiDi algorithm. Valid
                      values are 'L' or 'R'. If None is given the default
                      auto-resolution will be used.
        """
        if self._display_order:
            return self._reorder(base_dir)
        else:
            return self

    def display_order(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BBoxOCRRecord':
        """
        Returns the OCR record in Unicode display order, i.e. ordered from left
        to right inside the line.

        Args:
            base_dir: An optional string defining the base direction (also
                      called paragraph direction) for the BiDi algorithm. Valid
                      values are 'L' or 'R'. If None is given the default
                      auto-resolution will be used.
        """
        if self._display_order:
            return self
        else:
            return self._reorder(base_dir)

    def _reorder(self, base_dir: Optional[Literal['L', 'R']] = None) -> 'BBoxOCRRecord':
        storage = bd.get_empty_storage()

        if base_dir not in ('L', 'R'):
            base_level = bd.get_base_level(self.prediction)
        else:
            base_level = {'L': 0, 'R': 1}[base_dir]

        storage['base_level'] = base_level
        storage['base_dir'] = ('L', 'R')[base_level]

        bd.get_embedding_levels(self.prediction, storage)
        bd.explicit_embed_and_overrides(storage)
        bd.resolve_weak_types(storage)
        bd.resolve_neutral_types(storage, False)
        bd.resolve_implicit_levels(storage, False)
        for i, j in enumerate(zip(self.prediction, self.cuts, self.confidences)):
            storage['chars'][i]['record'] = j
        bd.reorder_resolved_levels(storage, False)
        bd.apply_mirroring(storage, False)
        prediction = ''
        cuts = []
        confidences = []
        for ch in storage['chars']:
            # code point may have been mirrored
            prediction = prediction + ch['ch']
            cuts.append(ch['record'][1])
            confidences.append(ch['record'][2])
        # carry over whole line information
        line = BBoxLine(id=self.id,
                        bbox=self.bbox,
                        text=self.text,
                        base_dir=self._line_base_dir,
                        imagename=self.imagename,
                        tags=self.tags,
                        split=self.split,
                        regions=self.regions)
        rec = BBoxOCRRecord(prediction=prediction,
                            cuts=cuts,
                            confidences=confidences,
                            line=line,
                            base_dir=base_dir,
                            display_order=not self._display_order)
        return rec



# -*- coding: utf-8 -*-
import json
import os
import unittest
from io import BytesIO
from pathlib import Path

from lxml import etree
from PIL import Image

from kraken.transcribe import TranscriptionInterface

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestTranscriptionInterface(unittest.TestCase):

    """
    Test of the transcription interface generation
    """

    def test_transcription_generation(self):
        """
        Tests creation of transcription interfaces with segmentation.
        """
        tr = TranscriptionInterface()
        with open(resources / 'segmentation.json') as fp:
            seg = json.load(fp)
        with Image.open(resources / 'input.jpg') as im:
            tr.add_page(im, seg)
        fp = BytesIO()
        tr.write(fp)
        # this will not throw an exception ever so we need a better validator
        etree.HTML(fp.getvalue())



#
# Copyright 2021 Teklia
# Copyright 2021 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
align
~~~~~

A character alignment module using a network output lattice and ground truth to
accuractely determine grapheme locations in input data.
"""
import dataclasses
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Optional

import torch
from bidi.algorithm import get_display
from PIL import Image

from kraken import rpred
from kraken.containers import BaselineOCRRecord, Segmentation

if TYPE_CHECKING:
    from kraken.lib.models import TorchSeqRecognizer

logger = logging.getLogger('kraken')


def forced_align(doc: Segmentation, model: 'TorchSeqRecognizer', base_dir: Optional[Literal['L', 'R']] = None) -> Segmentation:
    """
    Performs a forced character alignment of text with recognition model
    output activations.

    Argument:
        doc: Parsed document.
        model: Recognition model to use for alignment.

    Returns:
        A Segmentation object where the record's contain the aligned text.
    """
    im = Image.open(doc.imagename)
    predictor = rpred.rpred(model, im, doc)

    records = []

    # enable training mode in last layer to get log_softmax output
    model.nn.nn[-1].training = True

    for idx, line in enumerate(doc.lines):
        # convert text to display order
        do_text = get_display(line.text, base_dir=base_dir)
        # encode into labels, ignoring unencodable sequences
        labels = model.codec.encode(do_text).long()
        next(predictor)
        if model.outputs.shape[2] < 2*len(labels):
            logger.warning(f'Could not align line {idx}. Output sequence length {model.outputs.shape[2]} < '
                           f'{2*len(labels)} (length of "{line.text}" after encoding).')
            records.append(BaselineOCRRecord('', [], [], line))
            continue
        emission = torch.tensor(model.outputs).squeeze().T
        trellis = get_trellis(emission, labels)
        path = backtrack(trellis, emission, labels)
        path = merge_repeats(path, do_text)
        pred = []
        pos = []
        conf = []
        for seg in path:
            pred.append(seg.label)
            pos.append((predictor._scale_val(seg.start, 0, predictor.box.size[0]),
                        predictor._scale_val(seg.end, 0, predictor.box.size[0])))
            conf.append(seg.score)
        records.append(BaselineOCRRecord(pred, pos, conf, line, display_order=True))
    return dataclasses.replace(doc, lines=records)


"""
Copied from the forced alignment with Wav2Vec2 tutorial of pytorch available
at:
https://github.com/pytorch/audio/blob/main/examples/tutorials/forced_alignment_tutorial.py
"""


@dataclass
class Point:
    token_index: int
    time_index: int
    score: float


# Merge the labels
@dataclass
class Segment:
    label: str
    start: int
    end: int
    score: float

    def __repr__(self):
        return f"{self.label}\t({self.score:4.2f}): [{self.start:5d}, {self.end:5d})"

    @property
    def length(self):
        return self.end - self.start


def get_trellis(emission, tokens):
    # width x labels in log domain
    num_frame = emission.size(0)
    num_tokens = len(tokens)

    # Trellis has extra dimensions for both time axis and tokens.
    # The extra dim for tokens represents <SoS> (start-of-sentence)
    # The extra dim for time axis is for simplification of the code.
    trellis = torch.empty((num_frame + 1, num_tokens + 1))
    trellis[0, 0] = 0
    trellis[1:, 0] = torch.cumsum(emission[:, 0], 0)
    trellis[0, -num_tokens:] = -float("inf")
    trellis[-num_tokens:, 0] = float("inf")

    for t in range(num_frame):
        trellis[t + 1, 1:] = torch.maximum(
            # Score for staying at the same token
            trellis[t, 1:] + emission[t, 0],
            # Score for changing to the next token
            trellis[t, :-1] + emission[t, tokens],
        )
    return trellis


def backtrack(trellis, emission, tokens):
    # Note:
    # j and t are indices for trellis, which has extra dimensions
    # for time and tokens at the beginning.
    # When referring to time frame index `T` in trellis,
    # the corresponding index in emission is `T-1`.
    # Similarly, when referring to token index `J` in trellis,
    # the corresponding index in transcript is `J-1`.
    j = trellis.size(1) - 1
    t_start = torch.argmax(trellis[:, j]).item()

    path = []
    for t in range(t_start, 0, -1):
        # 1. Figure out if the current position was stay or change
        # Note (again):
        # `emission[J-1]` is the emission at time frame `J` of trellis dimension.
        # Score for token staying the same from time frame J-1 to T.
        stayed = trellis[t - 1, j] + emission[t - 1, 0]
        # Score for token changing from C-1 at T-1 to J at T.
        changed = trellis[t - 1, j - 1] + emission[t - 1, tokens[j - 1]]

        # 2. Store the path with frame-wise probability.
        prob = emission[t - 1, tokens[j - 1] if changed > stayed else 0].exp().item()
        # Return token index and time index in non-trellis coordinate.
        path.append(Point(j - 1, t - 1, prob))

        # 3. Update the token
        if changed > stayed:
            j -= 1
            if j == 0:
                break
    else:
        raise ValueError("Failed to align")
    return path[::-1]


def merge_repeats(path, ground_truth):
    i1, i2 = 0, 0
    segments = []
    while i1 < len(path):
        while i2 < len(path) and path[i1].token_index == path[i2].token_index:
            i2 += 1
        score = sum(path[k].score for k in range(i1, i2)) / (i2 - i1)
        segments.append(
            Segment(
                ground_truth[path[i1].token_index],
                path[i1].time_index,
                path[i2 - 1].time_index + 1,
                score,
            )
        )
        i1 = i2
    return segments



# -*- coding: utf-8 -*-

from contextlib import contextmanager
import unittest
import tempfile
from unittest.mock import Mock, patch
from pathlib import Path
from traceback import print_exc
import warnings
from typing import Optional, List, Union

from PIL import Image

from click.testing import CliRunner

from kraken.containers import (
    BaselineLine,
    BaselineOCRRecord,
    BBoxLine,
    BBoxOCRRecord,
    Segmentation,
)
from kraken.lib import xml
from kraken.lib import segmentation
from kraken.lib.models import load_any
from kraken.rpred import mm_rpred, rpred
from kraken.kraken import cli as kraken_cli
from kraken.ketos import cli as ketos_cli
import re

thisfile = Path(__file__).resolve().parent
resources = thisfile / "resources"

def mock_extract_polygons():
    return Mock(side_effect=segmentation.extract_polygons)

class TestNewPolygons(unittest.TestCase):
    """
    Tests for the new polygon extraction method.
    """

    def setUp(self):
        self.im = Image.open(resources / "bw.png")
        self.old_model_path = str(resources / "overfit.mlmodel")
        self.old_model = load_any(self.old_model_path)
        self.new_model_path = str(resources / "overfit_newpoly.mlmodel")
        self.new_model = load_any(self.new_model_path)
        self.segmented_img = str(resources / "170025120000003,0074-lite.xml")
        self.runner = CliRunner()
        self.color_img = resources / "input.tif"
        self.arrow_data = str(resources / "merge_tests/base.arrow")
        self.simple_bl_seg = Segmentation(
            type="baselines",
            imagename=resources / "bw.png",
            lines=[
                BaselineLine(
                    id="foo",
                    baseline=[[0, 10], [2543, 10]],
                    boundary=[[0, 0], [2543, 0], [2543, 155], [0, 155]],
                )
            ],
            text_direction="horizontal-lr",
            script_detection=False,
        )

    ## RECIPES

    @patch("kraken.rpred.extract_polygons", new_callable=mock_extract_polygons)
    def _test_rpred(self, extractor_mock: Mock, *, model, force_no_legacy: bool=False, expect_legacy: bool):
        """
        Base recipe for testing rpred with a given model and polygon extraction method
        """
        pred = rpred(model, self.im, self.simple_bl_seg, True, no_legacy_polygons=force_no_legacy)
        _ = next(pred)

        extractor_mock.assert_called()
        for cl in extractor_mock.mock_calls:
            self.assertEqual(cl[2]["legacy"], expect_legacy)

    @patch("kraken.rpred.extract_polygons", new_callable=mock_extract_polygons)
    def _test_krakencli(self, extractor_mock: Mock, *, args, force_no_legacy: bool=False, expect_legacy: bool,):
        """
        Base recipe for testing kraken_cli with a given polygon extraction method
        """
        if force_no_legacy:
            args = ["--no-legacy-polygons"] + args

        result = self.runner.invoke(kraken_cli, args)
        print("kraken", *args)

        if result.exception:
            print_exc()

        self.assertEqual(result.exit_code, 0)
        extractor_mock.assert_called()
        for cl in extractor_mock.mock_calls:
            self.assertEqual(cl[2]["legacy"], expect_legacy)

    def _test_ketoscli(self, *, args, expect_legacy: bool, check_exit_code: Optional[Union[int, List[int]]] = 0, patching_dir="kraken.lib.dataset.recognition"):
        """
        Base recipe for testing ketos_cli with a given polygon extraction method
        """
        with patch(patching_dir + ".extract_polygons", new_callable=mock_extract_polygons) as extractor_mock:
            result = self.runner.invoke(ketos_cli, args)

            print("ketos", *args)
            if result.exception:
                print(result.output)
                print_exc()

            if check_exit_code is not None:
                if isinstance(check_exit_code, int):
                    check_exit_code = [check_exit_code]
                self.assertIn(result.exit_code, check_exit_code, "Command failed")

            extractor_mock.assert_called()
            for cl in extractor_mock.mock_calls:
                self.assertEqual(cl[2]["legacy"], expect_legacy)

    ## TESTS

    def test_rpred_from_old_model(self):
        """
        Test rpred with old model, check that it uses legacy polygon extraction method
        """
        self._test_rpred(model=self.old_model, force_no_legacy=False, expect_legacy=True)

    def test_rpred_from_old_model_force_new(self):
        """
        Test rpred with old model, but disabling legacy polygons
        """
        self._test_rpred(model=self.old_model, force_no_legacy=True, expect_legacy=False)

    def test_rpred_from_new_model(self):
        """
        Test rpred with new model, check that it uses new polygon extraction method
        """
        self._test_rpred(model=self.new_model, force_no_legacy=False, expect_legacy=False)


    def test_krakencli_ocr_old_model(self):
        """
        Test kraken_cli with old model, check that it uses legacy polygon extraction method
        """
        with tempfile.NamedTemporaryFile() as fp:
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp.name, 'ocr', '-m', self.old_model_path],
                force_no_legacy=False,
                expect_legacy=True,
            )

    def test_krakencli_ocr_old_model_force_new(self):
        """
        Test kraken_cli with old model, check that it uses legacy polygon extraction method
        """
        with tempfile.NamedTemporaryFile() as fp:
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp.name, 'ocr', '-m', self.old_model_path],
                force_no_legacy=True,
                expect_legacy=False,
            )

    def test_krakencli_ocr_new_model(self):
        """
        Test kraken_cli with new model, check that it uses new polygon extraction method
        """
        with tempfile.NamedTemporaryFile() as fp:
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp.name, 'ocr', '-m', self.new_model_path],
                force_no_legacy=False,
                expect_legacy=False,
            )


    def test_ketoscli_test_old_model(self):
        """
        Test `ketos test` with old model, check that it uses legacy polygon extraction method
        """
        self._test_ketoscli(
            args=['test', '-m', self.old_model_path, '-f', 'xml', '--workers', '0', self.segmented_img],
            expect_legacy=True,
        )

    def test_ketoscli_test_old_model_force_new(self):
        """
        Test `ketos test` with old model, check that it does not use legacy polygon extraction method
        """
        self._test_ketoscli(
            args=['test', '--no-legacy-polygons', '-m', self.old_model_path, '-f', 'xml', '--workers', '0', self.segmented_img],
            expect_legacy=False,
        )

    def test_ketoscli_test_new_model(self):
        """
        Test `ketos test` with new model, check that it uses new polygon extraction method
        """
        self._test_ketoscli(
            args=['test', '-m', self.new_model_path, '-f', 'xml', '--workers', '0', self.segmented_img],
            expect_legacy=False,
        )

    @unittest.skip('fails randomly')
    def test_ketoscli_train_new_model(self):
        """
        Test `ketos train` with new model, check that it uses new polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['train', '-f', 'xml', '-N', '1', '-q', 'fixed', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=False,
                check_exit_code=[0, 1], # Model may not improve during training
            )

            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=False,
            )

    @unittest.skip('fails randomly')
    def test_ketoscli_train_new_model_force_legacy(self):
        """
        Test `ketos train` training new model, check that it uses legacy polygon extraction method if forced
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['train', '--legacy-polygons', '-f', 'xml', '-N', '1', '-q', 'fixed', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=True,
                check_exit_code=[0, 1], # Model may not improve during training
            )

            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=True,
            )

    @unittest.skip('fails randomly')
    def test_ketoscli_train_old_model(self):
        """
        Test `ketos train` finetuning old model, check that it uses new polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['train', '-f', 'xml', '-N', '1', '-q', 'fixed', '-i', self.old_model_path, '--resize', 'add', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=False,
                check_exit_code=[0, 1], # Model may not improve during training
            )
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=False,
            )

    @unittest.skip('fails randomly')
    def test_ketoscli_train_old_model_force_legacy(self):
        """
        Test `ketos train` finetuning old model, check that it uses legacy polygon extraction method if forced
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['train', '--legacy-polygons', '-f', 'xml', '-N', '1', '-q', 'fixed', '-i', self.old_model_path, '--resize', 'add', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=True,
                check_exit_code=[0, 1], # Model may not improve during training
            )
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=True,
            )

    @unittest.expectedFailure
    def test_ketoscli_pretrain_new_model(self):
        """
        Test `ketos pretrain` with new model, check that it uses new polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['pretrain', '-f', 'xml', '-N', '1', '-q', 'fixed', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=False,
                check_exit_code=[0, 1], # Model may not improve during training
            )
            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=False,
            )

    @unittest.expectedFailure
    def test_ketoscli_pretrain_new_model_force_legacy(self):
        """
        Test `ketos pretrain` with new model, check that it uses legacy polygon extraction method if forced
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['pretrain', '--legacy-polygons', '-f', 'xml', '-N', '1', '-q', 'fixed', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=True,
                check_exit_code=[0, 1], # Model may not improve during training
            )

            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', str(mfp) + "_0.mlmodel"],
                expect_legacy=True,
            )

    @unittest.expectedFailure
    def test_ketoscli_pretrain_old_model(self):
        """
        Test `ketos pretrain` with old model, check that it uses new polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            fp = str(Path(tempdir) / "test.xml")

            self._test_ketoscli(
                args=['pretrain', '-f', 'xml', '-N', '1', '-q', 'fixed', '-i', self.old_model_path, '--resize', 'add', '-o', mfp, '--workers', '0', self.segmented_img],
                expect_legacy=False,
                check_exit_code=[0, 1], # Model may not improve during training
            )

            self._test_krakencli(
                args=['-f', 'xml', '-i', self.segmented_img, fp, 'ocr', '-m', mfp + "_0.mlmodel"],
                expect_legacy=False,
            )


    def _assertWarnsWhenTrainingArrow(self,
                                      model: str,
                                      *dset: str,
                                      from_model: Optional[str] = None,
                                      force_legacy: bool = False,
                                      expect_warning_msgs: List[str] = [],
                                      expect_not_warning_msgs: List[str] = []):

        args = ['-f', 'binary', '-N', '1', '-q', 'fixed', '-o', model, *dset]
        if force_legacy:
            args = ['--legacy-polygons'] + args
        if from_model:
            args = ['-i', from_model, '--resize', 'add'] + args

        print("ketos", 'train', *args)
        run = self.runner.invoke(ketos_cli, ['train'] + args)
        output = re.sub(r'\w+\.py:\d+\n', '', run.output)
        output = re.sub(r'\s+', ' ', output)
        for warning_msg in expect_warning_msgs:
            self.assertIn(warning_msg, output, f"Expected warning '{warning_msg}' not found in output")
        for warning_msg in expect_not_warning_msgs:
            self.assertNotIn(warning_msg, output, f"Unexpected warning '{warning_msg}' found in output")

    def test_ketos_old_arrow_train_new(self):
        """
        Test `ketos train`, on old arrow dataset, check that it raises a warning about polygon extraction method only if incoherent
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            mfp2 = str(Path(tempdir) / "model2")

            self._assertWarnsWhenTrainingArrow(mfp, self.arrow_data, force_legacy=False, expect_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set", "the new model will be flagged to use legacy"])
            self._assertWarnsWhenTrainingArrow(mfp2, self.arrow_data, force_legacy=True, expect_not_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set", "the new model will be flagged to use legacy"])

    def test_ketos_new_arrow(self):
        """
        Test `ketos compile`, check that it uses new polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            dset = str(Path(tempdir) / "dataset.arrow")
            mfp = str(Path(tempdir) / "model")
            mfp2 = str(Path(tempdir) / "model2")

            self._test_ketoscli(
                args=['compile', '--workers', '0', '-f', 'xml', '-o', dset, self.segmented_img],
                expect_legacy=False,
                patching_dir="kraken.lib.arrow_dataset",
            )

            self._assertWarnsWhenTrainingArrow(mfp, dset, force_legacy=False, expect_not_warning_msgs=["WARNING Setting dataset legacy polygon status to False based on training set", "the new model will be flagged to use legacy"])
            self._assertWarnsWhenTrainingArrow(mfp2, dset, force_legacy=True, expect_warning_msgs=["WARNING Setting dataset legacy polygon status to False based on training set", "the new model will be flagged to use new"])


    def test_ketos_new_arrow_force_legacy(self):
        """
        Test `ketos compile`, check that it uses old polygon extraction method
        """
        with tempfile.TemporaryDirectory() as tempdir:
            dset = str(Path(tempdir) / "dataset.arrow")
            mfp = str(Path(tempdir) / "model")
            mfp2 = str(Path(tempdir) / "model2")

            self._test_ketoscli(
                args=['compile', '--workers', '0', '--legacy-polygons', '-f', 'xml', '-o', dset, self.segmented_img],
                expect_legacy=True,
                patching_dir="kraken.lib.arrow_dataset",
            )

            self._assertWarnsWhenTrainingArrow(mfp, dset, force_legacy=False, expect_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set", "the new model will be flagged to use legacy"])
            self._assertWarnsWhenTrainingArrow(mfp2, dset, force_legacy=True, expect_not_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set", "the new model will be flagged to use legacy"])

    def test_ketos_old_arrow_old_model(self):
        """
        Test `ketos train`, on old arrow dataset, check that it raises a warning about polygon extraction method only if incoherent
        """
        with tempfile.TemporaryDirectory() as tempdir:
            mfp = str(Path(tempdir) / "model")
            mfp2 = str(Path(tempdir) / "model2")

            self._assertWarnsWhenTrainingArrow(mfp, self.arrow_data, from_model=self.old_model_path, force_legacy=False, expect_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set"], expect_not_warning_msgs=["model will be flagged to use new"])
            self._assertWarnsWhenTrainingArrow(mfp2, self.arrow_data, from_model=self.old_model_path, force_legacy=True, expect_not_warning_msgs=["WARNING Setting dataset legacy polygon status to True based on training set", "model will be flagged to use new"])

    def test_ketos_new_arrow_old_model(self):
        """
        Test `ketos compile`, on new arrow dataset, check that it raises a warning about polygon extraction method only if incoherent
        """
        with tempfile.TemporaryDirectory() as tempdir:
            dset = str(Path(tempdir) / "dataset.arrow")
            mfp = str(Path(tempdir) / "model")
            mfp2 = str(Path(tempdir) / "model2")

            self._test_ketoscli(
                args=['compile', '--workers', '0', '-f', 'xml', '-o', dset, self.segmented_img],
                expect_legacy=False,
                patching_dir="kraken.lib.arrow_dataset",
            )

            self._assertWarnsWhenTrainingArrow(mfp, dset, from_model=self.old_model_path, force_legacy=False, expect_not_warning_msgs=["WARNING Setting dataset legacy polygon status to False based on training set"], expect_warning_msgs=["model will be flagged to use new"])
            self._assertWarnsWhenTrainingArrow(mfp2, dset, from_model=self.old_model_path, force_legacy=True, expect_warning_msgs=["WARNING Setting dataset legacy polygon status to False based on training set"], expect_not_warning_msgs=["model will be flagged to use new"])

    def test_ketos_mixed_arrow_train_new(self):
        """
        Test `ketos compile`, on mixed arrow dataset, check that it raises a warning about polygon extraction method only if incoherent
        """
        with tempfile.TemporaryDirectory() as tempdir:
            dset = str(Path(tempdir) / "dataset.arrow")
            mfp = str(Path(tempdir) / "model")

            self._test_ketoscli(
                args=['compile', '--workers', '0', '-f', 'xml', '-o', dset, self.segmented_img, self.arrow_data],
                expect_legacy=False,
                patching_dir="kraken.lib.arrow_dataset",
            )

            self._assertWarnsWhenTrainingArrow(mfp, dset, self.arrow_data, force_legacy=True, expect_warning_msgs=["WARNING Mixed legacy polygon", "WARNING Setting dataset legacy polygon status to False based on training set"], expect_not_warning_msgs=["model will be flagged to use legacy"])



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.train
~~~~~~~~~~~~~~~~~~

Command line driver for recognition training and evaluation.
"""
import logging
import pathlib
from typing import List
from functools import partial
import warnings

import click
from threadpoolctl import threadpool_limits

from kraken.lib.default_specs import RECOGNITION_HYPER_PARAMS, RECOGNITION_SPEC
from kraken.lib.exceptions import KrakenInputException

from .util import _expand_gt, _validate_manifests, message, to_ptl_device

logging.captureWarnings(True)
logger = logging.getLogger('kraken')


@click.command('train')
@click.pass_context
@click.option('-B', '--batch-size', show_default=True, type=click.INT,
              default=RECOGNITION_HYPER_PARAMS['batch_size'], help='batch sample size')
@click.option('--pad', show_default=True, type=click.INT, default=16, help='Left and right '
              'padding around lines')
@click.option('-o', '--output', show_default=True, type=click.Path(), default='model', help='Output model file')
@click.option('-s', '--spec', show_default=True, default=RECOGNITION_SPEC,
              help='VGSL spec of the network to train. CTC layer will be added automatically.')
@click.option('-a', '--append', show_default=True, default=None, type=click.INT,
              help='Removes layers before argument and then appends spec. Only works when loading an existing model')
@click.option('-i', '--load', show_default=True, type=click.Path(exists=True,
              readable=True), help='Load existing file to continue training')
@click.option('-F', '--freq', show_default=True, default=RECOGNITION_HYPER_PARAMS['freq'], type=click.FLOAT,
              help='Model saving and report generation frequency in epochs '
                   'during training. If frequency is >1 it must be an integer, '
                   'i.e. running validation every n-th epoch.')
@click.option('-q',
              '--quit',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['quit'],
              type=click.Choice(['early',
                                 'fixed']),
              help='Stop condition for training. Set to `early` for early stooping or `fixed` for fixed number of epochs')
@click.option('-N',
              '--epochs',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['epochs'],
              help='Number of epochs to train for')
@click.option('--min-epochs',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['min_epochs'],
              help='Minimal number of epochs to train for when using early stopping.')
@click.option('--lag',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['lag'],
              help='Number of evaluations (--report frequency) to wait before stopping training without improvement')
@click.option('--min-delta',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['min_delta'],
              type=click.FLOAT,
              help='Minimum improvement between epochs to reset early stopping. Default is scales the delta by the best loss')
@click.option('-d', '--device', show_default=True, default='cpu', help='Select device to use (cpu, cuda:0, cuda:1, ...)')
@click.option('--precision',
              show_default=True,
              default='32',
              type=click.Choice(['64', '32', 'bf16', '16']),
              help='Numerical precision to use for training. Default is 32-bit single-point precision.')
@click.option('--optimizer',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['optimizer'],
              type=click.Choice(['Adam',
                                 'SGD',
                                 'RMSprop']),
              help='Select optimizer')
@click.option('-r', '--lrate', show_default=True, default=RECOGNITION_HYPER_PARAMS['lrate'], help='Learning rate')
@click.option('-m', '--momentum', show_default=True, default=RECOGNITION_HYPER_PARAMS['momentum'], help='Momentum')
@click.option('-w', '--weight-decay', show_default=True, type=float,
              default=RECOGNITION_HYPER_PARAMS['weight_decay'], help='Weight decay')
@click.option('--warmup', show_default=True, type=int,
              default=RECOGNITION_HYPER_PARAMS['warmup'], help='Number of steps to ramp up to `lrate` initial learning rate.')
@click.option('--freeze-backbone', show_default=True, type=int,
              default=RECOGNITION_HYPER_PARAMS['freeze_backbone'], help='Number of samples to keep the backbone (everything but last layer) frozen.')
@click.option('--schedule',
              show_default=True,
              type=click.Choice(['constant',
                                 '1cycle',
                                 'exponential',
                                 'cosine',
                                 'step',
                                 'reduceonplateau']),
              default=RECOGNITION_HYPER_PARAMS['schedule'],
              help='Set learning rate scheduler. For 1cycle, cycle length is determined by the `--epoch` option.')
@click.option('-g',
              '--gamma',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['gamma'],
              help='Decay factor for exponential, step, and reduceonplateau learning rate schedules')
@click.option('-ss',
              '--step-size',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['step_size'],
              help='Number of validation runs between learning rate decay for exponential and step LR schedules')
@click.option('--sched-patience',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['rop_patience'],
              help='Minimal number of validation runs between LR reduction for reduceonplateau LR schedule.')
@click.option('--cos-max',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['cos_t_max'],
              help='Epoch of minimal learning rate for cosine LR scheduler.')
@click.option('--cos-min-lr',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['cos_min_lr'],
              help='Minimal final learning rate for cosine LR scheduler.')
@click.option('-p', '--partition', show_default=True, default=0.9,
              help='Ground truth data partition ratio between train/validation set')
@click.option('--fixed-splits/--ignore-fixed-split', show_default=True, default=False,
              help='Whether to honor fixed splits in binary datasets.')
@click.option('-u', '--normalization', show_default=True, type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']),
              default=RECOGNITION_HYPER_PARAMS['normalization'], help='Ground truth normalization')
@click.option('-n', '--normalize-whitespace/--no-normalize-whitespace', show_default=True,
              default=RECOGNITION_HYPER_PARAMS['normalize_whitespace'], help='Normalizes unicode whitespace')
@click.option('-c', '--codec', show_default=True, default=None, type=click.File(mode='r', lazy=True),
              help='Load a codec JSON definition (invalid if loading existing model)')
@click.option('--resize', show_default=True, default='fail',
              type=click.Choice([
                  'add', 'union',  # Deprecation: `add` is deprecated, `union` is the new value
                  'both', 'new',  # Deprecation: `both` is deprecated, `new` is the new value
                  'fail'
              ]),
              help='Codec/output layer resizing option. If set to `union` code '
                   'points will be added, `new` will set the layer to match exactly '
                   'the training data, `fail` will abort if training data and model '
                   'codec do not match.')
@click.option('--reorder/--no-reorder', show_default=True, default=True, help='Reordering of code points to display order')
@click.option('--base-dir', show_default=True, default='auto',
              type=click.Choice(['L', 'R', 'auto']), help='Set base text '
              'direction.  This should be set to the direction used during the '
              'creation of the training data. If set to `auto` it will be '
              'overridden by any explicit value given in the input files.')
@click.option('-t', '--training-files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with additional paths to training data')
@click.option('-e', '--evaluation-files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with paths to evaluation data. Overrides the `-p` parameter')
@click.option('--workers', show_default=True, default=1, type=click.IntRange(0), help='Number of worker processes.')
@click.option('--threads', show_default=True, default=1, type=click.IntRange(1), help='Maximum size of OpenMP/BLAS thread pool.')
@click.option('--load-hyper-parameters/--no-load-hyper-parameters', show_default=True, default=False,
              help='When loading an existing model, retrieve hyperparameters from the model')
@click.option('--force-binarization/--no-binarization', show_default=True,
              default=False, help='Forces input images to be binary, otherwise '
              'the appropriate color format will be auto-determined through the '
              'network specification. Will be ignored in `path` mode.')
@click.option('-f', '--format-type', type=click.Choice(['path', 'xml', 'alto', 'page', 'binary']), default='path',
              help='Sets the training data format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both line definitions and a '
              'link to source images. In `path` mode arguments are image files '
              'sharing a prefix up to the last extension with `.gt.txt` text files '
              'containing the transcription. In binary mode files are datasets '
              'files containing pre-extracted text lines.')
@click.option('--augment/--no-augment',
              show_default=True,
              default=RECOGNITION_HYPER_PARAMS['augment'],
              help='Enable image augmentation')
@click.option('--logger', 'pl_logger', show_default=True, type=click.Choice(['tensorboard']), default=None,
              help='Logger used by PyTorch Lightning to track metrics such as loss and accuracy.')
@click.option('--log-dir', show_default=True, type=click.Path(exists=True, dir_okay=True, writable=True),
              help='Path to directory where the logger will store the logs. If not set, a directory will be created in the current working directory.')
@click.argument('ground_truth', nargs=-1, callback=_expand_gt, type=click.Path(exists=False, dir_okay=False))
@click.option('--legacy-polygons', show_default=True, default=False, is_flag=True, help='Use the legacy polygon extractor.')
def train(ctx, batch_size, pad, output, spec, append, load, freq, quit, epochs,
          min_epochs, lag, min_delta, device, precision, optimizer, lrate,
          momentum, weight_decay, warmup, freeze_backbone, schedule, gamma,
          step_size, sched_patience, cos_max, cos_min_lr, partition,
          fixed_splits, normalization, normalize_whitespace, codec, resize,
          reorder, base_dir, training_files, evaluation_files, workers,
          threads, load_hyper_parameters, force_binarization,
          format_type, augment, pl_logger, log_dir, ground_truth,
          legacy_polygons):
    """
    Trains a model from image-text pairs.
    """
    if not load and append:
        raise click.BadOptionUsage('append', 'append option requires loading an existing model')

    if resize != 'fail' and not load:
        raise click.BadOptionUsage('resize', 'resize option requires loading an existing model')

    if not (0 <= freq <= 1) and freq % 1.0 != 0:
        raise click.BadOptionUsage('freq', 'freq needs to be either in the interval [0,1.0] or a positive integer.')

    if augment:
        try:
            import albumentations  # NOQA
        except ImportError:
            raise click.BadOptionUsage('augment', 'augmentation needs the `albumentations` package installed.')

    if pl_logger == 'tensorboard':
        try:
            import tensorboard  # NOQA
        except ImportError:
            raise click.BadOptionUsage('logger', 'tensorboard logger needs the `tensorboard` package installed.')

    if log_dir is None:
        log_dir = pathlib.Path.cwd()

    import json
    import shutil

    from kraken.lib.train import KrakenTrainer, RecognitionModel

    hyper_params = RECOGNITION_HYPER_PARAMS.copy()
    hyper_params.update({'freq': freq,
                         'pad': pad,
                         'batch_size': batch_size,
                         'quit': quit,
                         'epochs': epochs,
                         'min_epochs': min_epochs,
                         'lag': lag,
                         'min_delta': min_delta,
                         'optimizer': optimizer,
                         'lrate': lrate,
                         'momentum': momentum,
                         'weight_decay': weight_decay,
                         'warmup': warmup,
                         'freeze_backbone': freeze_backbone,
                         'schedule': schedule,
                         'gamma': gamma,
                         'step_size': step_size,
                         'rop_patience': sched_patience,
                         'cos_t_max': cos_max,
                         'cos_min_lr': cos_min_lr,
                         'normalization': normalization,
                         'normalize_whitespace': normalize_whitespace,
                         'augment': augment,
                         })

    # disable automatic partition when given evaluation set explicitly
    if evaluation_files:
        partition = 1
    ground_truth = list(ground_truth)

    # merge training_files into ground_truth list
    if training_files:
        ground_truth.extend(training_files)

    if len(ground_truth) == 0:
        raise click.UsageError('No training data was provided to the train command. Use `-t` or the `ground_truth` argument.')

    if reorder and base_dir != 'auto':
        reorder = base_dir

    if codec:
        logger.debug(f'Loading codec file from {codec}')
        codec = json.load(codec)

    try:
        accelerator, device = to_ptl_device(device)
    except Exception as e:
        raise click.BadOptionUsage('device', str(e))

    if hyper_params['freq'] > 1:
        val_check_interval = {'check_val_every_n_epoch': int(hyper_params['freq'])}
    else:
        val_check_interval = {'val_check_interval': hyper_params['freq']}

    model = RecognitionModel(hyper_params=hyper_params,
                             output=output,
                             spec=spec,
                             append=append,
                             model=load,
                             reorder=reorder,
                             training_data=ground_truth,
                             evaluation_data=evaluation_files,
                             partition=partition,
                             binary_dataset_split=fixed_splits,
                             num_workers=workers,
                             load_hyper_parameters=load_hyper_parameters,
                             force_binarization=force_binarization,
                             format_type=format_type,
                             codec=codec,
                             resize=resize,
                             legacy_polygons=legacy_polygons)

    # Force upgrade to new polygon extractor if model was not trained with it
    if model.nn and model.nn.use_legacy_polygons:
        if not legacy_polygons and not model.legacy_polygons:
            # upgrade to new polygon extractor
            logger.warning('The model will be flagged to use new polygon extractor.')
            model.nn.use_legacy_polygons = False
    if not model.nn and legacy_polygons != model.legacy_polygons:
        logger.warning(f'Dataset was compiled with legacy polygon extractor: {model.legacy_polygons}, '
                       f'the new model will be flagged to use {"legacy" if model.legacy_polygons else "new"} method.')
        legacy_polygons = model.legacy_polygons

    trainer = KrakenTrainer(accelerator=accelerator,
                            devices=device,
                            precision=precision,
                            max_epochs=hyper_params['epochs'] if hyper_params['quit'] == 'fixed' else -1,
                            min_epochs=hyper_params['min_epochs'],
                            freeze_backbone=hyper_params['freeze_backbone'],
                            enable_progress_bar=True if not ctx.meta['verbose'] else False,
                            deterministic=ctx.meta['deterministic'],
                            pl_logger=pl_logger,
                            log_dir=log_dir,
                            **val_check_interval)
    try:
        with threadpool_limits(limits=threads):
            trainer.fit(model)
    except KrakenInputException as e:
        if e.args[0].startswith('Training data and model codec alphabets mismatch') and resize == 'fail':
            raise click.BadOptionUsage('resize', 'Mismatched training data for loaded model. Set option `--resize` to `new` or `add`')
        else:
            raise e

    if model.best_epoch == -1:
        logger.warning('Model did not improve during training.')
        ctx.exit(1)

    if not model.current_epoch:
        logger.warning('Training aborted before end of first epoch.')
        ctx.exit(1)

    if quit == 'early':
        message(f'Moving best model {model.best_model} ({model.best_metric}) to {output}_best.mlmodel')
        logger.info(f'Moving best model {model.best_model} ({model.best_metric}) to {output}_best.mlmodel')
        shutil.copy(f'{model.best_model}', f'{output}_best.mlmodel')


@click.command('test')
@click.pass_context
@click.option('-B', '--batch-size', show_default=True, type=click.INT,
              default=RECOGNITION_HYPER_PARAMS['batch_size'], help='Batch sample size')
@click.option('-m', '--model', show_default=True, type=click.Path(exists=True, readable=True),
              multiple=True, help='Model(s) to evaluate')
@click.option('-e', '--evaluation-files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with paths to evaluation data.')
@click.option('-d', '--device', show_default=True, default='cpu', help='Select device to use (cpu, cuda:0, cuda:1, ...)')
@click.option('--pad', show_default=True, type=click.INT, default=16, help='Left and right '
              'padding around lines')
@click.option('--workers', show_default=True, default=1,
              type=click.IntRange(0),
              help='Number of worker processes when running on CPU.')
@click.option('--threads', show_default=True, default=1,
              type=click.IntRange(1),
              help='Max size of thread pools for OpenMP/BLAS operations.')
@click.option('--reorder/--no-reorder', show_default=True, default=True, help='Reordering of code points to display order')
@click.option('--base-dir', show_default=True, default='auto',
              type=click.Choice(['L', 'R', 'auto']), help='Set base text '
              'direction.  This should be set to the direction used during the '
              'creation of the training data. If set to `auto` it will be '
              'overridden by any explicit value given in the input files.')
@click.option('-u', '--normalization', show_default=True, type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']),
              default=None, help='Ground truth normalization')
@click.option('-n', '--normalize-whitespace/--no-normalize-whitespace',
              show_default=True, default=True, help='Normalizes unicode whitespace')
@click.option('--force-binarization/--no-binarization', show_default=True,
              default=False, help='Forces input images to be binary, otherwise '
              'the appropriate color format will be auto-determined through the '
              'network specification. Will be ignored in `path` mode.')
@click.option('-f', '--format-type', type=click.Choice(['path', 'xml', 'alto', 'page', 'binary']), default='path',
              help='Sets the training data format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both baselines and a '
              'link to source images. In `path` mode arguments are image files '
              'sharing a prefix up to the last extension with JSON `.path` files '
              'containing the baseline information. In `binary` mode files are '
              'collections of pre-extracted text line images.')
@click.option('--fixed-splits/--ignore-fixed-split', show_default=True, default=False,
              help='Whether to honor fixed splits in binary datasets.')
@click.argument('test_set', nargs=-1, callback=_expand_gt, type=click.Path(exists=False, dir_okay=False))
@click.option('--no-legacy-polygons', show_default=True, default=False, is_flag=True, help='Force disable the legacy polygon extractor.')
def test(ctx, batch_size, model, evaluation_files, device, pad, workers,
         threads, reorder, base_dir, normalization, normalize_whitespace,
         force_binarization, format_type, fixed_splits, test_set, no_legacy_polygons):
    """
    Evaluate on a test set.
    """
    if not model:
        raise click.UsageError('No model to evaluate given.')

    import numpy as np
    from torch.utils.data import DataLoader

    from torchmetrics.text import CharErrorRate, WordErrorRate

    from kraken.lib import models, util
    from kraken.lib.dataset import (ArrowIPCRecognitionDataset,
                                    GroundTruthDataset, ImageInputTransforms,
                                    PolygonGTDataset, collate_sequences,
                                    compute_confusions, global_align)
    from kraken.lib.progress import KrakenProgressBar
    from kraken.lib.xml import XMLPage
    from kraken.serialization import render_report

    logger.info('Building test set from {} line images'.format(len(test_set) + len(evaluation_files)))

    legacy_polygons = None
    incoherent_legacy_polygons = False

    nn = {}
    for p in model:
        message('Loading model {}\t'.format(p), nl=False)
        nn[p] = models.load_any(p, device)
        message('\u2713', fg='green')
        model_legacy_polygons = nn[p].nn.use_legacy_polygons
        if legacy_polygons is None:
            legacy_polygons = model_legacy_polygons
        elif legacy_polygons != model_legacy_polygons:
            incoherent_legacy_polygons = True

    if incoherent_legacy_polygons and not no_legacy_polygons:
        logger.warning('Models use different polygon extractors. Legacy polygon extractor will be used ; use --no-legacy-polygons to force disable it.')
        legacy_polygons = True
    elif no_legacy_polygons:
        legacy_polygons = False

    if legacy_polygons:
        warnings.warn('Using legacy polygon extractor, as the model was not trained with the new method. Please retrain your model to get performance improvements.')

    pin_ds_mem = False
    if device != 'cpu':
        pin_ds_mem = True

    test_set = list(test_set)

    if evaluation_files:
        test_set.extend(evaluation_files)

    if len(test_set) == 0:
        raise click.UsageError('No evaluation data was provided to the test command. Use `-e` or the `test_set` argument.')

    dataset_kwargs = {}
    if fixed_splits:
        if format_type != "binary":
            logger.warning("--fixed-splits can only be use with data using binary format")
        else:
            dataset_kwargs["split_filter"] = "test"

    if format_type in ['xml', 'page', 'alto']:
        test_set = [{'page': XMLPage(file, filetype=format_type).to_container()} for file in test_set]
        valid_norm = False
        DatasetClass = partial(PolygonGTDataset, legacy_polygons=legacy_polygons)
    elif format_type == 'binary':
        DatasetClass = ArrowIPCRecognitionDataset
        test_set = [{'file': file} for file in test_set]
        valid_norm = False
    else:
        DatasetClass = GroundTruthDataset
        if force_binarization:
            logger.warning('Forced binarization enabled in `path` mode. Will be ignored.')
            force_binarization = False
        test_set = [{'line': util.parse_gt_path(img)} for img in test_set]
        valid_norm = True

    if len(test_set) == 0:
        raise click.UsageError('No evaluation data was provided to the test command. Use `-e` or the `test_set` argument.')

    if reorder and base_dir != 'auto':
        reorder = base_dir

    cer_list = []
    wer_list = []

    with threadpool_limits(limits=threads):
        for p, net in nn.items():
            algn_gt: List[str] = []
            algn_pred: List[str] = []
            chars = 0
            error = 0
            message('Evaluating {}'.format(p))
            logger.info('Evaluating {}'.format(p))
            batch, channels, height, width = net.nn.input
            ts = ImageInputTransforms(batch, height, width, channels, (pad, 0), valid_norm, force_binarization)
            ds = DatasetClass(normalization=normalization,
                              whitespace_normalization=normalize_whitespace,
                              reorder=reorder,
                              im_transforms=ts,
                              **dataset_kwargs)
            for line in test_set:
                try:
                    ds.add(**line)
                except ValueError as e:
                    logger.info(e)

            if hasattr(ds, 'legacy_polygon_status'):
                if ds.legacy_polygons_status != legacy_polygons:
                    warnings.warn(
                        f'Binary dataset was compiled with legacy polygon extractor: {ds.legacy_polygon_status}, '
                        f'while expecting data extracted with {"legacy" if legacy_polygons else "new"} method. Results may be inaccurate.')

            # don't encode validation set as the alphabets may not match causing encoding failures
            ds.no_encode()
            ds_loader = DataLoader(ds,
                                   batch_size=batch_size,
                                   num_workers=workers,
                                   pin_memory=pin_ds_mem,
                                   collate_fn=collate_sequences)

            test_cer = CharErrorRate()
            test_wer = WordErrorRate()

            with KrakenProgressBar() as progress:
                batches = len(ds_loader)
                pred_task = progress.add_task('Evaluating', total=batches, visible=True if not ctx.meta['verbose'] else False)

                for batch in ds_loader:
                    im = batch['image']
                    text = batch['target']
                    lens = batch['seq_lens']
                    try:
                        pred = net.predict_string(im, lens)
                        for x, y in zip(pred, text):
                            chars += len(y)
                            c, algn1, algn2 = global_align(y, x)
                            algn_gt.extend(algn1)
                            algn_pred.extend(algn2)
                            error += c
                            test_cer.update(x, y)
                            test_wer.update(x, y)

                    except FileNotFoundError as e:
                        batches -= 1
                        progress.update(pred_task, total=batches)
                        logger.warning('{} {}. Skipping.'.format(e.strerror, e.filename))
                    except KrakenInputException as e:
                        batches -= 1
                        progress.update(pred_task, total=batches)
                        logger.warning(str(e))
                    progress.update(pred_task, advance=1)

            cer_list.append(1.0 - test_cer.compute())
            wer_list.append(1.0 - test_wer.compute())
            confusions, scripts, ins, dels, subs = compute_confusions(algn_gt, algn_pred)
            rep = render_report(p,
                                chars,
                                error,
                                cer_list[-1],
                                wer_list[-1],
                                confusions,
                                scripts,
                                ins,
                                dels,
                                subs)
            logger.info(rep)
            message(rep)

    logger.info('Average character accuracy: {:0.2f}%, (stddev: {:0.2f})'.format(np.mean(cer_list) * 100, np.std(cer_list) * 100))
    message('Average character accuracy: {:0.2f}%, (stddev: {:0.2f})'.format(np.mean(cer_list) * 100, np.std(cer_list) * 100))
    logger.info('Average word accuracy: {:0.2f}%, (stddev: {:0.2f})'.format(np.mean(wer_list) * 100, np.std(wer_list) * 100))
    message('Average word accuracy: {:0.2f}%, (stddev: {:0.2f})'.format(np.mean(wer_list) * 100, np.std(wer_list) * 100))



#
# Copyright 2017 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Pytorch compatible codec with many-to-many mapping between labels and
graphemes.
"""
import logging
from collections import Counter
from typing import Dict, List, Sequence, Set, Tuple, Union

import numpy as np
from torch import IntTensor

from kraken.lib.exceptions import KrakenCodecException, KrakenEncodeException

__all__ = ['PytorchCodec']

logger = logging.getLogger(__name__)


class PytorchCodec(object):
    """
    Builds a codec converting between graphemes/code points and integer
    label sequences.

    charset may either be a string, a list or a dict. In the first case
    each code point will be assigned a label, in the second case each
    string in the list will be assigned a label, and in the final case each
    key string will be mapped to the value sequence of integers. In the
    first two cases labels will be assigned automatically. When a mapping
    is manually provided the label codes need to be a prefix-free code.

    As 0 is the blank label in a CTC output layer, output labels and input
    dictionaries are/should be 1-indexed.

    Args:
        charset: Input character set.
        strict: Flag indicating if encoding/decoding errors should be ignored
                or cause an exception.

    Raises:
        KrakenCodecException: If the character set contains duplicate
                              entries or the mapping is non-singular or
                              non-prefix-free.
    """
    def __init__(self, charset: Union[Dict[str, Sequence[int]], Sequence[str], str], strict=False):
        if isinstance(charset, dict):
            self.c2l = charset
        else:
            cc = Counter(charset)
            if len(cc) < len(charset):
                raise KrakenCodecException(f'Duplicate entry in codec definition string: {cc}')
            self.c2l = {k: [v] for v, k in enumerate(sorted(charset), start=1)}
        self.c_sorted = sorted(self.c2l.keys(), key=len, reverse=True)
        self.l2c: Dict[Tuple[int], str] = {tuple(v): k for k, v in self.c2l.items()}
        self.l2c_single = {k[0]: v for k, v in self.l2c.items() if len(k) == 1}
        self.strict = strict
        if not self.is_valid:
            raise KrakenCodecException('Codec is not valid (non-singular/non-prefix free).')

    def __len__(self) -> int:
        """
        Total number of input labels the codec can decode.
        """
        return len(self.l2c.keys())

    @property
    def is_valid(self) -> bool:
        """
        Returns True if the codec is prefix-free (in label space) and
        non-singular (in both directions).
        """
        # quick test for non-singularity
        if len(self.l2c.keys()) != len(self.c2l.keys()):
            return False

        for i, code_1 in enumerate(sorted(self.l2c.keys())):
            for j, code_2 in enumerate(sorted(self.l2c.keys())):
                if i != j and code_1[:len(code_2)] == code_2:
                    return False

        return True

    @property
    def max_label(self) -> int:
        """
        Returns the maximum label value.
        """
        return max(label for labels in self.c2l.values() for label in labels)

    def encode(self, s: str) -> IntTensor:
        """
        Encodes a string into a sequence of labels.

        If the code is non-singular we greedily encode the longest sequence first.

        Args:
            s: Input unicode string

        Returns:
            Ecoded label sequence

        Raises:
            KrakenEncodeException: if the a subsequence is not encodable and the
                                   codec is set to strict mode.
        """
        labels: List[int] = []
        idx = 0
        while idx < len(s):
            encodable_suffix = False
            for code in self.c_sorted:
                if len(code) == 1:
                    break
                if s[idx:].startswith(code):
                    labels.extend(self.c2l[code])
                    idx += len(code)
                    encodable_suffix = True
                    break

            if not encodable_suffix and s[idx] in self.c2l:
                labels.extend(self.c2l[s[idx]])
                idx += 1
                encodable_suffix = True

            if not encodable_suffix:
                if self.strict:
                    raise KrakenEncodeException(f'Non-encodable sequence {s[idx:idx+5]}... encountered.')
                logger.warning(f'Non-encodable sequence {s[idx:idx+5]}... encountered. Advancing one code point.')
                idx += 1

        return IntTensor(labels)

    def decode(self, labels: Sequence[Tuple[int, int, int, float]]) -> List[Tuple[str, int, int, float]]:
        """
        Decodes a labelling.

        Given a labelling with cuts and  confidences returns a string with the
        cuts and confidences aggregated across label-code point
        correspondences. When decoding multilabels to code points the resulting
        cuts are min/max, confidences are averaged.

        Args:
            labels: Input containing tuples (label, start, end,
                           confidence).

        Returns:
            A list of tuples (code point, start, end, confidence)
        """
        start = [x for _, x, _, _ in labels]
        end = [x for _, _, x, _ in labels]
        con = [x for _, _, _, x in labels]
        labels = tuple(x for x, _, _, _ in labels)
        decoded = []
        idx = 0
        while idx < len(labels):
            decodable_suffix = False
            if int(labels[idx]) in self.l2c_single:
                code = self.l2c_single[int(labels[idx])]
                decoded.extend([(c, s, e, u) for c, s, e, u in zip(code,
                                                                   len(code) * [start[idx]],
                                                                   len(code) * [end[idx]],
                                                                   len(code) * [con[idx]])])
                idx += 1
                decodable_suffix = True
            else:
                for code in self.l2c.keys():
                    if code == labels[idx:idx+len(code)]:
                        decoded.extend([(c, s, e, u) for c, s, e, u in zip(self.l2c[code],
                                                                           len(self.l2c[code]) * [start[idx]],
                                                                           len(self.l2c[code]) * [end[idx + len(code) - 1]],
                                                                           len(self.l2c[code]) * [np.mean(con[idx:idx + len(code)])])])
                        idx += len(code)
                        decodable_suffix = True
                        break
            if not decodable_suffix:
                if self.strict:
                    raise KrakenEncodeException(f'Non-decodable sequence {labels[idx:idx+5]}... encountered.')
                logger.debug(f'Non-decodable sequence {labels[idx:idx+5]}... encountered. Advancing one label.')
                idx += 1
        return decoded

    def merge(self, codec: 'PytorchCodec') -> Tuple['PytorchCodec', Set]:
        """
        Transforms this codec (c1) into another (c2) reusing as many labels as
        possible.

        The resulting codec is able to encode the same code point sequences
        while not necessarily having the same labels for them as c2.
        Retains matching character -> label mappings from both codecs, removes
        mappings not c2, and adds mappings not in c1. Compound labels in c2 for
        code point sequences not in c1 containing labels also in use in c1 are
        added as separate labels.

        Args:
            codec: PytorchCodec to merge with

        Returns:
            A merged codec and a list of labels that were removed from the
            original codec.
        """
        # find character sequences not encodable (exact match) by new codec.
        # get labels for these sequences as deletion candidates
        rm_candidates = {cseq: enc for cseq, enc in self.c2l.items() if cseq not in codec.c2l}
        c2l_cand = self.c2l.copy()
        for x in rm_candidates.keys():
            c2l_cand.pop(x)
        # remove labels from candidate list that are in use for other decodings
        rm_labels = [label for v in rm_candidates.values() for label in v]
        for v in c2l_cand.values():
            for label in rm_labels:
                if label in v:
                    rm_labels.remove(label)
        # iteratively remove labels, decrementing subsequent labels to close
        # (new) holes in the codec.
        offset_rm_labels = [v-idx for idx, v in enumerate(sorted(set(rm_labels)))]
        for rlabel in offset_rm_labels:
            c2l_cand = {k: [label-1 if label > rlabel else label for label in v] for k, v in c2l_cand.items()}
        # add mappings not in original codec
        add_list = {cseq: enc for cseq, enc in codec.c2l.items() if cseq not in self.c2l}
        # renumber
        start_idx = max((0,) + tuple(label for v in c2l_cand.values() for label in v)) + 1
        add_labels = {k: v for v, k in enumerate(sorted(set(label for v in add_list.values() for label in v)), start_idx)}
        for k, v in add_list.items():
            c2l_cand[k] = [add_labels[label] for label in v]
        return PytorchCodec(c2l_cand, self.strict), set(rm_labels)

    def add_labels(self, charset: Union[Dict[str, Sequence[int]], Sequence[str], str]) -> 'PytorchCodec':
        """
        Adds additional characters/labels to the codec.

        charset may either be a string, a list or a dict. In the first case
        each code point will be assigned a label, in the second case each
        string in the list will be assigned a label, and in the final case each
        key string will be mapped to the value sequence of integers. In the
        first two cases labels will be assigned automatically.

        As 0 is the blank label in a CTC output layer, output labels and input
        dictionaries are/should be 1-indexed.

        Args:
            charset: Input character set.
        """
        if isinstance(charset, dict):
            c2l = self.c2l.copy()
            c2l.update(charset)
        else:
            c2l = self.c2l.copy()
            c2l.update({k: [v] for v, k in enumerate(sorted(charset), start=self.max_label+1)})
        return PytorchCodec(c2l, self.strict)

    def __repr__(self):
        return f'PytorchCodec({self.c2l})'



# -*- coding: utf-8 -*-
import shutil
import tempfile
import unittest
from pathlib import Path

from kraken import repo

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestRepo(unittest.TestCase):
    """
    Testing interaction with the model repository.
    """

    def setUp(self):
        self.temp_model = tempfile.TemporaryDirectory()
        self.temp_path = Path(self.temp_model.name)

    def tearDown(self):
        shutil.rmtree(self.temp_model.name)

    def test_listing(self):
        """
        Tests fetching the model list.
        """
        records = repo.get_listing()
        self.assertGreater(len(records), 15)

    def test_get_description(self):
        """
        Tests fetching the description of a model.
        """
        record = repo.get_description('10.5281/zenodo.8425684')
        self.assertEqual(record['doi'], '10.5281/zenodo.8425684')

    def test_get_model(self):
        """
        Tests fetching a model.
        """
        id = repo.get_model('10.5281/zenodo.8425684',
                            path=self.temp_model.name)
        self.assertEqual(id, 'omnisyr_best.mlmodel')
        self.assertEqual((self.temp_path / id).stat().st_size, 16245671)

    def test_prev_record_version_get_description(self):
        """
        Tests fetching the description of a model that has a superseding newer version.
        """
        record = repo.get_description('10.5281/zenodo.6657809')
        self.assertEqual(record['doi'], '10.5281/zenodo.6657809')

    def test_prev_record_version_get_model(self):
        """
        Tests fetching a model that has a superseding newer version.
        """
        id = repo.get_model('10.5281/zenodo.6657809',
                            path=self.temp_model.name)
        self.assertEqual(id, 'HTR-United-Manu_McFrench.mlmodel')
        self.assertEqual((self.temp_path / id).stat().st_size, 16176844)

